/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.ast;

import java.util.*;

import polyglot.ast.*;
import polyglot.types.Flags;
import polyglot.util.*;

import sj.ast.protocoldecls.*;
import sj.ast.typenodes.*;
import sj.ast.sessvars.*;
import sj.ast.createops.*;
import sj.ast.sessops.*;
import sj.ast.sessops.basicops.*;
import sj.ast.sessops.structuralops.*;
import sj.ast.socketops.*;
import sj.ast.servops.*;
import sj.ast.sesscasts.*;
import sj.ast.sessformals.*;
import sj.ast.proceduredecls.*;
import sj.util.*;

/**
 * NodeFactory for sj2 extension.
 */
public interface SJNodeFactory extends NodeFactory
{
	public SJFieldProtocolDecl SJFieldProtocolDecl(Position pos, Flags flags, Id name, SJTypeNode sjTypeNode);
	public SJLocalProtocolDecl SJLocalProtocolDecl(Position pos, Flags flags, Id name, SJTypeNode sjTypeNode);

	public SJBeginNode SJBeginNode(Position pos);
	public SJSendNode SJSendNode(Position pos, TypeNode messageType);
	public SJReceiveNode SJReceiveNode(Position pos, TypeNode messageType);
	public SJOutbranchNode SJOutbranchNode(Position pos, List<SJBranchCaseNode> branchCases);
	public SJInbranchNode SJInbranchNode(Position pos, List<SJBranchCaseNode> branchCases);
	public SJBranchCaseNode SJBranchCaseNode(Position pos, SJLabel lab, SJTypeNode sjTypeNode);
	public SJOutwhileNode SJOutwhileNode(Position pos, SJTypeNode body);
	public SJInwhileNode SJInwhileNode(Position pos, SJTypeNode body);
	public SJRecursionNode SJRecursionNode(Position pos, SJLabel lab, SJTypeNode body);
	public SJRecurseNode SJRecurseNode(Position pos, SJLabel lab);
	public SJProtocolNode SJProtocolNode(Position pos, String protocolName);

	public SJChannelCreate SJChannelCreate(Position pos, Receiver target, List<Expr> arguments);
	public SJSocketCreate SJSocketCreate(Position pos, Receiver target, List<Expr> arguments);
	public SJServerCreate SJServerCreate(Position pos, Receiver target, List<Expr> arguments);

	public SJSessionTry SJSessionTry(Position pos, Block tryBlock, List<Catch> catchBlocks, List<Receiver> sjSockets);
	public SJSessionTry SJSessionTry(Position pos, Block tryBlock, List<Catch> catchBlocks, Block finallyBlock, List<Receiver> sjSockets);

	public SJLocalSocket SJLocalSocket(Position pos, Id id);
	public SJLocalServer SJLocalServer(Position pos, Id id);

	public SJRequest SJRequest(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);
	public SJSend SJSend(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);
	public SJSendBoolean SJSendBoolean(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);
	public SJSendInt SJSendInt(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);
	public SJSendByte SJSendByte(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);
	public SJSendSession SJSendSession(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);
	public SJSendChannel SJSendChannel(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);

	public SJReceive SJReceive(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);
	public SJReceiveBoolean SJReceiveBoolean(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);
	public SJReceiveInt SJReceiveInt(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);
	public SJReceiveByte SJReceiveByte(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);
	public SJReceiveSession SJReceiveSession(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);
	public SJReceiveChannel SJReceiveChannel(Position pos, SJSocketVariable target, Id name, List<Expr> arguments);
	public SJRecurse SJRecurse(Position pos, SJSocketVariable target, Id name, List<Expr> arguments, SJLabel lab);

	public SJOutbranch SJOutbranch(Position pos, List<Stmt> statements, SJLabel lab, Receiver target);
	public SJInbranch SJInbranch(Position pos, List<SJInbranchCase> branchCases, Receiver target, List<Expr> arguments); // Make (potential) timeout arguments uniform across all input operations.
	public SJInbranchCase SJInbranchCase(Position pos, List<Stmt> statements, SJLabel lab);
	public SJOutwhile SJOutwhile(Position pos, Expr cond, Stmt body, Receiver target);
	public SJInwhile SJInwhile(Position pos, Expr cond, Stmt body, Receiver target);
	public SJRecursion SJRecursion(Position pos, Stmt body, SJLabel lab, Receiver target);

	public SJOutlabel SJOutlabel(Position pos, SJSocketVariable target, SJLabel lab);
	public SJInlabel SJInlabel(Position pos, SJSocketVariable target, List<Expr> arguments);
	public SJOutsync SJOutsync(Position pos, SJSocketVariable target, Expr cond);
	public SJInsync SJInsync(Position pos, SJSocketVariable target, List<Expr> arguments);
	public SJRecursionFalse SJRecursionFalse(Position pos, SJSocketVariable target, SJLabel lab);
	public SJRecursionTrue SJRecursionTrue(Position pos, SJSocketVariable target, SJLabel lab);

	public SJAccept SJAccept(Position pos, SJServerVariable target, Id name, List<Expr> arguments);

	public SJTypeCast SJTypeCast(Position pos, Expr expr, SJTypeNode sjTypeNode);
	public SJTypeFormal SJTypeFormal(Position pos, Flags flags, Id name, SJTypeNode sjTypeNode);

	public SJMethodDecl SJMethodDecl(Position pos, Flags flags, TypeNode returnType, Id name, List<Formal> formals, List<TypeNode> throwTypes, Block body);
	public SJMethodDecl SJMethodDecl(Position pos, Flags flags, SJTypeNode sjReturnType, Id name, List<Formal> formals, List<TypeNode> throwTypes, Block body);

	public SJRecursiveMethodDecl SJRecursiveMethodDecl(Position pos, Flags flags, TypeNode returnType, Id name, List<Formal> formals, List<TypeNode> throwTypes, Block body);
	public SJRecursiveMethodDecl SJRecursiveMethodDecl(Position pos, Flags flags, SJTypeNode sjReturnType, Id name, List<Formal> formals, List<TypeNode> throwTypes, Block body);

	public SJCall SJCall(Position pos, Receiver target, Id name, List<Expr> arguments);
	public SJReturn SJReturn(Position pos, Expr expr);
	public SJSpawn SJSpawn(Position pos, New target, List<Receiver> sockets);
}
