/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.ast;

import java.util.*;

import polyglot.ast.*;
import polyglot.types.Flags;
import polyglot.util.*;

import sj.SJConstants;
import sj.ast.protocoldecls.*;
import sj.ast.typenodes.*;
import sj.ast.sessvars.*;
import sj.ast.createops.*;
import sj.ast.sessops.*;
import sj.ast.sessops.basicops.*;
import sj.ast.sessops.structuralops.*;
import sj.ast.socketops.*;
import sj.ast.servops.*;
import sj.ast.sesscasts.*;
import sj.ast.sessformals.*;
import sj.ast.proceduredecls.*;
import sj.del.*;
import sj.extension.*;
import sj.extension.sesscasts.*;
import sj.util.*;

import static sj.SJConstants.*;

/**
 * NodeFactory for sj2 extension.
 */
public class SJNodeFactory_c extends NodeFactory_c implements SJNodeFactory
{
	public SJNodeFactory_c()
	{
		super(new SJExtFactory_c(), new SJDelFactory_c()); // Override extFactory method to return the specific type.
	}

	public SJExtFactory extFactory()
	{
		return (SJExtFactory) super.extFactory();
	}

	public SJDelFactory delFactory()
	{
		return (SJDelFactory) super.delFactory();
	}

	public SJFieldProtocolDecl SJFieldProtocolDecl(Position pos, Flags flags, Id name, SJTypeNode sjTypeNode)
	{
		if (!flags.isFinal()) // To utilise base type checking.
		{
			flags = flags.Final();
		}

		SJFieldProtocolDecl n = new SJFieldProtocolDecl_c(pos, flags, CanonicalTypeNode(pos, SJConstants.SJPROTOCOL_TYPE), name, NullLit(pos)); // Null initialization overwritten by protocol declaration translation pass (dummy init. needed to satsify base type checking).

		return (SJFieldProtocolDecl) n.ext(extFactory().SJProtocolDeclExt(sjTypeNode));
	}

	public SJLocalProtocolDecl SJLocalProtocolDecl(Position pos, Flags flags, Id name, SJTypeNode sjTypeNode)
	{
		if (!flags.isFinal()) // To utilise base type checking.
		{
			flags = flags.Final();
		}

		SJLocalProtocolDecl n = new SJLocalProtocolDecl_c(pos, flags, CanonicalTypeNode(pos, SJConstants.SJPROTOCOL_TYPE), name, NullLit(pos));

		return (SJLocalProtocolDecl) n.ext(extFactory().SJProtocolDeclExt(sjTypeNode));
	}

	public SJBeginNode SJBeginNode(Position pos)
	{
		return new SJBeginNode_c(pos);
	}

	public SJSendNode SJSendNode(Position pos, TypeNode messageType)
	{
		return new SJSendNode_c(pos, messageType);
	}

	public SJReceiveNode SJReceiveNode(Position pos, TypeNode messageType)
	{
		return new SJReceiveNode_c(pos, messageType);
	}

	public SJOutbranchNode SJOutbranchNode(Position pos, List<SJBranchCaseNode> branchCases)
	{
		return new SJOutbranchNode_c(pos, branchCases);
	}

	public SJInbranchNode SJInbranchNode(Position pos, List<SJBranchCaseNode> branchCases)
	{
		return new SJInbranchNode_c(pos, branchCases);
	}

	public SJBranchCaseNode SJBranchCaseNode(Position pos, SJLabel lab, SJTypeNode sjTypeNode)
	{
		return new SJBranchCaseNode_c(pos, lab, sjTypeNode);
	}

	public SJOutwhileNode SJOutwhileNode(Position pos, SJTypeNode body)
	{
		return new SJOutwhileNode_c(pos, body);
	}

	public SJInwhileNode SJInwhileNode(Position pos, SJTypeNode body)
	{
		return new SJInwhileNode_c(pos, body);
	}

	public SJRecursionNode SJRecursionNode(Position pos, SJLabel lab, SJTypeNode body)
	{
		return new SJRecursionNode_c(pos, lab, body);
	}

	public SJRecurseNode SJRecurseNode(Position pos, SJLabel lab)
	{
		return new SJRecurseNode_c(pos, lab);
	}

	public SJProtocolNode SJProtocolNode(Position pos, String protocolName)
	{
		return new SJProtocolNode_c(pos, protocolName);
	}

	public SJChannelCreate SJChannelCreate(Position pos, Receiver target, List<Expr> arguments)
	{
		Id name = Id(pos, SJCHANNEL_CREATE);

		SJChannelCreate n = new SJChannelCreate_c(pos, target, name, arguments);

		return (SJChannelCreate) n.ext(extFactory().SJCreateOperationExt());
	}

	public SJSocketCreate SJSocketCreate(Position pos, Receiver target, List<Expr> arguments)
	{
		Id name = Id(pos, SJSOCKET_CREATE);

		SJSocketCreate n = new SJSocketCreate_c(pos, target, name, arguments);

		return (SJSocketCreate) n.ext(extFactory().SJCreateOperationExt());
	}

	public SJServerCreate SJServerCreate(Position pos, Receiver target, List<Expr> arguments)
	{
		Id name = Id(pos, SJSERVER_CREATE);

		SJServerCreate n = new SJServerCreate_c(pos, target, name, arguments);

		return (SJServerCreate) n.ext(extFactory().SJCreateOperationExt());
	}

	public SJSessionTry SJSessionTry(Position pos, Block tryBlock, List<Catch> catchBlocks, List<Receiver> sockets)
	{
		return SJSessionTry(pos, tryBlock, catchBlocks, null, sockets);
	}

	public SJSessionTry SJSessionTry(Position pos, Block tryBlock, List<Catch> catchBlocks, Block finallyBlock, List<Receiver> sockets)
	{
		SJSessionTry n = new SJSessionTry_c(pos, tryBlock, catchBlocks, finallyBlock, sockets);

		n = (SJSessionTry) n.del(delFactory().SJSessionTryDel());
		n = (SJSessionTry) n.ext(extFactory().SJSessionTryExt());

		return n;
	}

	public SJLocalSocket SJLocalSocket(Position pos, Id id)
	{
		SJLocalSocket n = new SJLocalSocket_c(pos, id);

		return (SJLocalSocket) n.ext(extFactory().SJNamedExt()); // SJTypeable component unused.
	}

	public SJLocalServer SJLocalServer(Position pos, Id id)
	{
		SJLocalServer n = new SJLocalServer_c(pos, id);

		return (SJLocalServer) n.ext(extFactory().SJNamedExt());
	}

	public SJRequest SJRequest(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJRequest n = new SJRequest_c(pos, target, name, arguments);

		return (SJRequest) n.ext(extFactory().SJBasicOperationExt());
	}

	public SJSend SJSend(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJSend n = new SJSend_c(pos, target, name, arguments);

		return (SJSend) n.ext(extFactory().SJBasicOperationExt());
	}

	public SJSendBoolean SJSendBoolean(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJSendBoolean n = new SJSendBoolean_c(pos, target, name, arguments);

		return (SJSendBoolean) n.ext(extFactory().SJBasicOperationExt());
	}

	public SJSendInt SJSendInt(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJSendInt n = new SJSendInt_c(pos, target, name, arguments);

		return (SJSendInt) n.ext(extFactory().SJBasicOperationExt());
	}

	public SJSendByte SJSendByte(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJSendByte n = new SJSendByte_c(pos, target, name, arguments);

		return (SJSendByte) n.ext(extFactory().SJBasicOperationExt());
	}

	public SJSendSession SJSendSession(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJSendSession n = new SJSendSession_c(pos, target, name, arguments);

		return (SJSendSession) n.ext(extFactory().SJBasicOperationExt());
	}

	public SJSendChannel SJSendChannel(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJSendChannel n = new SJSendChannel_c(pos, target, name, arguments);

		return (SJSendChannel) n.ext(extFactory().SJBasicOperationExt());
	}

	public SJReceive SJReceive(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJReceive n = new SJReceive_c(pos, target, name, arguments);

		return (SJReceive) n.ext(extFactory().SJBasicOperationExt());
	}

	public SJReceiveBoolean SJReceiveBoolean(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJReceiveBoolean n = new SJReceiveBoolean_c(pos, target, name, arguments);

		return (SJReceiveBoolean) n.ext(extFactory().SJBasicOperationExt());
	}

	public SJReceiveInt SJReceiveInt(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJReceiveInt n = new SJReceiveInt_c(pos, target, name, arguments);

		return (SJReceiveInt) n.ext(extFactory().SJBasicOperationExt());
	}

	public SJReceiveByte SJReceiveByte(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJReceiveByte n = new SJReceiveByte_c(pos, target, name, arguments);

		return (SJReceiveByte) n.ext(extFactory().SJBasicOperationExt());
	}

	public SJReceiveSession SJReceiveSession(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJReceiveSession n = new SJReceiveSession_c(pos, target, name, arguments);

		Ext ext = extFactory().SJBasicOperationExt(); // The operation type.
		ext = ext.ext(extFactory().SJReturnOperationExt()); // Session returned type.

		return (SJReceiveSession) n.ext(ext);
	}

	public SJReceiveChannel SJReceiveChannel(Position pos, SJSocketVariable target, Id name, List<Expr> arguments)
	{
		SJReceiveChannel n = new SJReceiveChannel_c(pos, target, name, arguments);

		Ext ext = extFactory().SJBasicOperationExt();
		ext = ext.ext(extFactory().SJReturnOperationExt());

		return (SJReceiveChannel) n.ext(ext);
	}

	public SJRecurse SJRecurse(Position pos, SJSocketVariable target, Id name, List<Expr> arguments, SJLabel lab)
	{
		SJRecurse n = new SJRecurse_c(pos, target, name, arguments, lab);

		Ext ext = extFactory().SJBasicOperationExt();

		return (SJRecurse) n.ext(ext);
	}

	public SJOutbranch SJOutbranch(Position pos, List<Stmt> statements, SJLabel lab, Receiver target)
	{
		SJOutbranch n = new SJOutbranch_c(pos, statements, lab);

		n = (SJOutbranch) n.del(delFactory().SJStructuralOperationDel());
		n = (SJOutbranch) n.ext(extFactory().SJStructuralOperationExt(target));

		return n;
	}

	public SJInbranch SJInbranch(Position pos, List<SJInbranchCase> branchCases, Receiver target, List<Expr> arguments)
	{
		SJInbranch n = new SJInbranch_c(pos, branchCases, arguments); // Need to make (potential)  timeout arguments uniform for all input operations.

		n = (SJInbranch) n.del(delFactory().SJStructuralOperationDel());
		n = (SJInbranch) n.ext(extFactory().SJStructuralOperationExt(target));

		return n;
	}

	public SJInbranchCase SJInbranchCase(Position pos, List<Stmt> statements, SJLabel lab)
	{
		return new SJInbranchCase_c(pos, statements, lab);
	}

	public SJOutwhile SJOutwhile(Position pos, Expr cond, Stmt body, Receiver target)
	{
		SJOutwhile n = new SJOutwhile_c(pos, cond, body);

		n = (SJOutwhile) n.del(delFactory().SJStructuralOperationDel());
		n = (SJOutwhile) n.ext(extFactory().SJStructuralOperationExt(target));

		return n;
	}

	public SJInwhile SJInwhile(Position pos, Expr cond, Stmt body, Receiver target)
	{
		SJInwhile n = new SJInwhile_c(pos, cond, body);

		n = (SJInwhile) n.del(delFactory().SJStructuralOperationDel());
		n = (SJInwhile) n.ext(extFactory().SJStructuralOperationExt(target));

		return n;
	}

	public SJRecursion SJRecursion(Position pos, Stmt body, SJLabel lab, Receiver target)
	{
		List<Stmt> stmts = new LinkedList<Stmt>();

		stmts.add(body);

		SJRecursion n = new SJRecursion_c(pos, stmts, lab);

		n = (SJRecursion) n.del(delFactory().SJStructuralOperationDel());
		n = (SJRecursion) n.ext(extFactory().SJStructuralOperationExt(target)); // Not actually used for recursion: it does not need an auxiliary SJSocket runtime operation.

		return n;
	}

	public SJOutlabel SJOutlabel(Position pos, SJSocketVariable target, SJLabel lab)
	{
		Id name = Id(pos, SJConstants.SJSOCKET_OUTLABEL);
		List<Expr> arguments = new LinkedList<Expr>();

		arguments.add(StringLit(pos, lab.toString())); // Can't move into node constuctor (super call must come first, no easy access to node factory and can't modify Polyglot immutable argument list).

		return new SJOutlabel_c(pos, target, name, arguments);
	}

	public SJInlabel SJInlabel(Position pos, SJSocketVariable target, List<Expr> arguments)
	{
		Id name = Id(pos, SJConstants.SJSOCKET_INLABEL);
		//List<Expr> arguments = new LinkedList();

		return new SJInlabel_c(pos, target, name, arguments);
	}

	public SJOutsync SJOutsync(Position pos, SJSocketVariable target, Expr cond)
	{
		Id name = Id(pos, SJConstants.SJSOCKET_OUTSYNC);
		List<Expr> arguments = new LinkedList<Expr>();

		arguments.add(cond);

		return new SJOutsync_c(pos, target, name, arguments);
	}

	public SJInsync SJInsync(Position pos, SJSocketVariable target, List<Expr> arguments)
	{
		Id name = Id(pos, SJConstants.SJSOCKET_INSYNC);
		//List<Expr> arguments = new LinkedList();

		return new SJInsync_c(pos, target, name, arguments);
	}

	public SJRecursionFalse SJRecursionFalse(Position pos, SJSocketVariable target, SJLabel lab)
	{
		Id name = Id(pos, SJConstants.SJSOCKET_RECURSIONFALSE);
		List<Expr> arguments = new LinkedList<Expr>();

		arguments.add(StringLit(pos, lab.labelValue()));

		return new SJRecursionFalse_c(pos, target, name, arguments);
	}

	public SJRecursionTrue SJRecursionTrue(Position pos, SJSocketVariable target, SJLabel lab)
	{
		Id name = Id(pos, SJConstants.SJSOCKET_RECURSIONTRUE);
		List<Expr> arguments = new LinkedList<Expr>();

		arguments.add(StringLit(pos, lab.labelValue()));

		return new SJRecursionTrue_c(pos, target, name, arguments);
	}

	public SJAccept SJAccept(Position pos, SJServerVariable target, Id name, List<Expr> arguments)
	{
		SJAccept n = new SJAccept_c(pos, target, name, arguments);

		Ext ext = extFactory().SJServerOperationExt();
		ext = ext.ext(extFactory().SJReturnOperationExt());

		return (SJAccept) n.ext(ext);
	}

	public LocalDecl LocalDecl(Position pos, Flags flags, TypeNode type, Id name, Expr init)
	{
		LocalDecl n = new LocalDecl_c(pos, flags, type, name, init);

		//n = (LocalDecl)n.ext(extFactory.extLocalDecl());
		n = (LocalDecl) n.del(delFactory().SJReceiveDel());

		return n;
	}

	public LocalAssign LocalAssign(Position pos, Local left, Assign.Operator op, Expr right)
	{
		LocalAssign n = new LocalAssign_c(pos, left, op, right);

		//n = (LocalAssign)n.ext(extFactory.extLocalAssign());
		n = (LocalAssign) n.del(delFactory().SJReceiveDel());

		return n;
	}

	public Call Call(Position pos, Receiver target, Id name, List args)
	{
		Call n = new Call_c(pos, target, name, CollectionUtil.nonNullList(args));

		n = (Call) n.ext(extFactory().extCall());
		n = (Call) n.del(delFactory().SJReceiveDel()); // Can ignore the original delCall call.

		return n;
	}

	public SJTypeCast SJTypeCast(Position pos, Expr expr, SJTypeNode sjTypeNode)
	{
		SJTypeCast n;

		if (sjTypeNode instanceof SJBeginNode || (sjTypeNode instanceof SJProtocolNode && expr instanceof Call && ((Call) expr).name().equals(SJ_SYNTAX_RECEIVECHANNEL))) // Doesn't work for SJProtocolNode. // Hacked in check for channel-receive, but still broken for plain receive operations.
		{
			n = new SJTypeCast_c(pos, CanonicalTypeNode(pos, SJConstants.SJCHANNEL_TYPE), expr);
		}
		else
		{
			n = new SJTypeCast_c(pos, CanonicalTypeNode(pos, SJConstants.SJSOCKET_TYPE), expr);
		}

		n = (SJTypeCast) n.ext(extFactory().SJTypeCastExt());
		n = (SJTypeCast) ((SJTypeCastExt) n.ext()).sjTypeNode(sjTypeNode);

		return n;
	}

	public SJTypeFormal SJTypeFormal(Position pos, Flags flags, Id name, SJTypeNode sjTypeNode)
	{
		SJTypeFormal n = new SJTypeFormal_c(pos, flags, CanonicalTypeNode(pos, SJConstants.SJSOCKET_TYPE), name);

		return (SJTypeFormal) n.ext(extFactory().SJTypeFormalExt(sjTypeNode));
	}

	public SJMethodDecl SJMethodDecl(Position pos, Flags flags, TypeNode returnType, Id name, List<Formal> formals, List<TypeNode> throwTypes, Block body)
	{
		SJMethodDecl n = new SJMethodDecl_c(pos, flags, returnType, name, formals, throwTypes, body);

		return (SJMethodDecl) n.ext(extFactory().SJMethodDeclExt()); // Null return session type.
	}

	public SJMethodDecl SJMethodDecl(Position pos, Flags flags, SJTypeNode sjReturnType, Id name, List<Formal> formals, List<TypeNode> throwTypes, Block body)
	{
		SJMethodDecl n = new SJMethodDecl_c(pos, flags, CanonicalTypeNode(pos, SJConstants.SJSOCKET_TYPE), name, formals, throwTypes, body); // SJSocket base return type.

		n = (SJMethodDecl) n.ext(extFactory().SJMethodDeclExt());
		n = ((SJMethodDeclExt) n.ext()).sjReturnType(sjReturnType);

		return n;
	}

	public SJRecursiveMethodDecl SJRecursiveMethodDecl(Position pos, Flags flags, TypeNode returnType, Id name, List<Formal> formals, List<TypeNode> throwTypes, Block body) // void return type.
	{
		SJRecursiveMethodDecl n = new SJRecursiveMethodDecl_c(pos, flags, returnType, name, formals, throwTypes, body);

		return (SJRecursiveMethodDecl) n.ext(extFactory().SJMethodDeclExt());
	}

	public SJRecursiveMethodDecl SJRecursiveMethodDecl(Position pos, Flags flags, SJTypeNode sjReturnType, Id name, List<Formal> formals, List<TypeNode> throwTypes, Block body)
	{
		SJRecursiveMethodDecl n = new SJRecursiveMethodDecl_c(pos, flags, CanonicalTypeNode(pos, SJConstants.SJSOCKET_TYPE), name, formals, throwTypes, body);

		n = (SJRecursiveMethodDecl) n.ext(extFactory().SJMethodDeclExt());
		n = (SJRecursiveMethodDecl) ((SJMethodDeclExt) n.ext()).sjReturnType(sjReturnType);

		return n;
	}

	public SJCall SJCall(Position pos, Receiver target, Id name, List<Expr> arguments)
	{
		SJCall n = new SJCall_c(pos, target, name, arguments);

		return (SJCall) n.ext(extFactory().SJCallExt());
	}

	public SJReturn SJReturn(Position pos, Expr expr)
	{
		SJReturn n = new SJReturn_c(pos, expr);

		return (SJReturn) n.ext(extFactory().SJNamedExt());
	}

	public SJSpawn SJSpawn(Position pos, New target, List<Receiver> sockets)
	{
		Id name = Id(pos, SJConstants.SJ_SPAWN_START);
		SJSpawn n = new SJSpawn_c(pos, target, name, new ArrayList<Expr>());

		/*List args = new ArrayList();

		args.addAll(target.arguments());
		args.addAll(sockets);

		target.arguments(args); */

		n = (SJSpawn) n.del(delFactory().SJSpawnDel());
		n = (SJSpawn) n.ext(extFactory().SJSpawnExt(sockets));

		return n;
	}
}
