/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.types;

import java.util.*;

import polyglot.types.*;
import polyglot.util.Position;

import sj.types.typeobjects.sesstypes.*;
import sj.types.typeobjects.typeableinstances.*;
import sj.util.*;

public class SJTypeSystem_c extends TypeSystem_c implements SJTypeSystem
{
	public SJBeginType SJBeginType()
	{
		return new SJBeginType_c(this);
	}

	public SJSendType SJSendType(Type messageType) throws SemanticException
	{
		return new SJSendType_c(this, messageType);
	}

	public SJReceiveType SJReceiveType(Type messageType) throws SemanticException
	{
		return new SJReceiveType_c(this, messageType);
	}

	public SJOutbranchType SJOutbranchType()
	{
		return new SJOutbranchType_c(this);
	}

	public SJInbranchType SJInbranchType()
	{
		return new SJInbranchType_c(this);
	}

	public SJOutwhileType SJOutwhileType()
	{
		return new SJOutwhileType_c(this);
	}

	public SJInwhileType SJInwhileType()
	{
		return new SJInwhileType_c(this);
	}

	public SJRecursionType SJRecursionType(SJLabel lab)
	{
		return new SJRecursionType_c(this, lab);
	}

	public SJRecurseType SJRecurseType(SJLabel lab)
	{
		return new SJRecurseType_c(this, lab);
	}

	public SJUnknownType SJUnknownType()
	{
		return new SJUnknownType(this);
	}

	public SJTypeableInstance SJTypeableInstance(Position pos, SJType sjType)
	{
		return new SJTypeableInstance_c(this, pos, sjType);
	}

	public SJNamedInstance SJNamedInstance(Position pos, SJType sjType, String name)
	{
		return new SJNamedInstance_c(this, pos, sjType, name);
	}

	public SJSessionTryInstance SJSessionTryInstance(Position pos, SJMap sockets)
	{
		return new SJSessionTryInstance_c(this, pos, sockets);
	}

	public SJSessionOpInstance SJSessionOpInstance(Position pos, SJType sjType, String name)
	{
		return new SJSessionOpInstance_c(this, pos, sjType, name);
	}

	public SJMethodInstance SJMethodInstance(String name, SJMap sjFormalTypes, SJType sjReturnType)
	{
		return new SJMethodInstance_c(name, sjFormalTypes, sjReturnType);
	}

	public SJRecursiveMethodInstance SJRecursiveMethodInstance(String name, SJMap sjFormalTypes, SJType sjReturnType)
	{
		return new SJRecursiveMethodInstance_c(name, sjFormalTypes, sjReturnType);
	}

	public SJThreadInstance SJThreadInstance(String name, List<SJType> runTypes)
	{
		return new SJThreadInstance_c(name, runTypes);
	}

	public boolean wellFormedRecursions(SJType sjType)
	{
		try
		{
			while (sjType != null)
			{
				Set<SJLabel> labs = new HashSet<SJLabel>();

				wellFormedRecursionsAux(sjType, labs, false); // Should always return false for well-formed types.

				sjType = sjType.child();
			}
		}
		catch (SemanticException se)
		{
			return false;
		}

		return true;
	}

	private boolean wellFormedRecursionsAux(SJType sjType, Set<SJLabel> labs, boolean hasRecurse) throws SemanticException
	{
		if (sjType == null)
		{
			return hasRecurse;
		}

		if (sjType instanceof SJOutbranchType)
		{
			SJOutbranchType sjobt = (SJOutbranchType) sjType;

			boolean b = hasRecurse;

			for (SJLabel lab : sjobt.labelSet())
			{
				if (wellFormedRecursionsAux(sjobt.branchCase(lab), labs, b) && !hasRecurse)
				{
					hasRecurse = true;
				}
			}
		}
		else if (sjType instanceof SJInbranchType)
		{
			SJInbranchType sjibt = (SJInbranchType) sjType;

			boolean b = hasRecurse;

			for (SJLabel lab : sjibt.labelSet())
			{
				if (wellFormedRecursionsAux(sjibt.branchCase(lab), labs, b) && !hasRecurse)
				{
					hasRecurse = true;
				}
			}
		}
		else if (sjType instanceof SJOutwhileType)
		{
			hasRecurse = wellFormedRecursionsAux(((SJOutwhileType) sjType).body(), new HashSet<SJLabel>(), hasRecurse);
		}
		else if (sjType instanceof SJInwhileType)
		{
			hasRecurse = wellFormedRecursionsAux(((SJInwhileType) sjType).body(), new HashSet<SJLabel>(), hasRecurse);
		}
		else if (sjType instanceof SJRecursionType)
		{
			SJRecursionType sjrt = (SJRecursionType) sjType;

			Set<SJLabel> newLabs = new HashSet<SJLabel>();

			newLabs.addAll(labs);
			newLabs.add(sjrt.label());

			hasRecurse = wellFormedRecursionsAux(sjrt.body(), newLabs, hasRecurse);

			if (labs.isEmpty())
			{
				hasRecurse = false;
			}
		}
		else if (sjType instanceof SJRecurseType)
		{
			SJRecurseType sjrt = (SJRecurseType) sjType;

			if (!labs.contains(sjrt.label()))
			{
				throw new SemanticException("Recurse label not bound: " + sjrt);
			}

			hasRecurse = true;
		}

		if (sjType.child() != null)
		{
			if (hasRecurse)
			{
				throw new SemanticException("Only tail recursion is permitted: " + sjType);
			}

			return wellFormedRecursionsAux(sjType.child(), labs, hasRecurse);
		}
		else
		{
			return hasRecurse;
		}
	}
}
