/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.types.typeobjects.sesstypes;

import java.util.*;

import polyglot.types.*;

import sj.util.SJLabel;

import static sj.SJConstants.*;

public class SJInbranchType_c extends SJType_c implements SJInbranchType
{
	public static final long serialVersionUID = 1L;

	private HashMap<SJLabel, SJType> cases;

	public SJInbranchType_c(TypeSystem ts)
	{
		super(ts);

		this.cases = new HashMap<SJLabel, SJType>();
	}

	public Set<SJLabel> labelSet()
	{
		return cases.keySet();
	}

	public boolean hasCase(SJLabel lab)
	{
		return labelSet().contains(lab);
	}

	public SJType branchCase(SJLabel lab)
	{
		SJType sjType = cases.get(lab);

		return (sjType == null) ? null : sjType.treeClone();
	}

	public SJInbranchType branchCase(SJLabel lab, SJType branchCase)
	{
		if (branchCase == null)
		{
			cases.put(lab, branchCase);
		}
		else
		{
			branchCase = branchCase.treeClone(); // Defensive copy.

			cases.put(lab, branchCase);

			((SJType_c) branchCase).parent(this);
		}

		return this;
	}

	public SJType removeCase(SJLabel lab)
	{
		return cases.remove(lab);
	}

	public boolean nodeEquals(SJType sjType)
	{
		if (sjType == null || !(sjType instanceof SJInbranchType))
		{
			return false;
		}

		SJInbranchType them = (SJInbranchType) sjType;
		Set<SJLabel> ourLabels = labelSet();
		Set<SJLabel> theirLabels = them.labelSet();

		if (!ourLabels.equals(theirLabels))
		{
			return false;
		}

		for (SJLabel lab : ourLabels)
		{
			SJType ourCase = branchCase(lab);
			SJType theirCase = them.branchCase(lab);

			if (ourCase == null)
			{
				if (theirCase != null) return false;
			}
			else
			{
				if (!ourCase.treeEquals(theirCase)) return false;
			}
		}

		return true;
	}

	public boolean nodeSubtype(SJType sjType)
	{
		if (sjType == null || !(sjType instanceof SJInbranchType))
		{
			return false;
		}

		SJInbranchType them = (SJInbranchType) sjType;
		Set<SJLabel> ourLabels = labelSet();
		Set<SJLabel> theirLabels = them.labelSet();

		if (!theirLabels.containsAll(ourLabels))
		{
			return false;
		}

		for (SJLabel lab : ourLabels)
		{
			SJType ourCase = branchCase(lab);
			SJType theirCase = them.branchCase(lab);

			if (ourCase == null)
			{
				if (theirCase != null) return false;
			}
			else
			{
				if (!ourCase.isSubtype(theirCase)) return false;
			}
		}

		return true;
	}

	public boolean nodeDualtype(SJType sjType)
	{
		if (sjType == null || !(sjType instanceof SJOutbranchType))
		{
			return false;
		}

		SJOutbranchType them = (SJOutbranchType) sjType;
		Set<SJLabel> ourLabels = labelSet();
		Set<SJLabel> theirLabels = them.labelSet();

		if (!ourLabels.containsAll(theirLabels))
		{
			return false;
		}

		for (SJLabel lab : theirLabels)
		{
			SJType ourCase = branchCase(lab);
			SJType theirCase = them.branchCase(lab);

			if (ourCase == null)
			{
				if (theirCase != null) return false;
			}
			else
			{
				if (!ourCase.isDualtype(theirCase)) return false;
			}
		}

		return true;
	}

	public SJType nodeSubsume(SJType sjType) throws SemanticException
	{
		if (!(sjType instanceof SJInbranchType))
		{
			throw new SemanticException("Not subsumable: " + this + ", " + sjType);
		}

		SJInbranchType them = (SJInbranchType) sjType;

		SJInbranchType res = typeSystem().SJInbranchType();

		for (SJLabel lab : labelSet())
		{
			if (them.hasCase(lab))
			{
				res.branchCase(lab, branchCase(lab).subsume(them.branchCase(lab)));
			}
		}

		return res;
	}

	public boolean nodeWellFormed()
	{
		if (labelSet().isEmpty())
		{
			return false;
		}

		for (SJLabel lab : labelSet())
		{
			SJType sjt = branchCase(lab);

			if (sjt != null && !sjt.treeWellFormed())
			{
				return false;
			}
		}

		return true;
	}

	public SJType nodeClone()
	{
		SJInbranchType sjibt = typeSystem().SJInbranchType();

		for (SJLabel lab : labelSet())
		{
			SJType branchCase = branchCase(lab);

			sjibt.branchCase(lab, (branchCase == null) ? null : (SJType) branchCase.treeClone());
		}

		return sjibt;
	}

	public String nodeToString()
	{
		String s = SJ_STRING_INBRANCH_OPEN;

		for (Iterator<SJLabel> i = labelSet().iterator(); i.hasNext(); )
		{
			SJLabel lab = i.next();
			SJType branchCase = branchCase(lab);

			s = s + lab + SJ_STRING_LABEL;

			if (branchCase == null)
			{
				s = s + " ";
			}
			else
			{
				s = s + branchCase(lab).toString();
			}

			if (i.hasNext()) s = s + SJ_STRING_CASE_SEPARATOR + " ";
		}

		return s + SJ_STRING_INBRANCH_CLOSE;
	}
}
