/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.visit.ContextVisitor;
import polyglot.visit.NodeVisitor;

import sj.SJConstants;
import sj.ast.*;
import sj.ast.createops.*;
import sj.ast.sessops.basicops.*;
import sj.ast.servops.*;
import sj.types.*;
import sj.types.typeobjects.typeableinstances.SJNamedInstance;
import sj.visit.contexts.*;

import sj.util.SJCompilerUtils;

// Aliasing via session argument passing is checked when the session-method call is parsed by the session operation parser. // Can't combine this class with SJCreateOperation parser because higher-order (receive) operations will not have been parsed. But should sort out exactly what each pass (SJCreateOperationParser, SJAliasChecker, ...) is responsible for.
public class SJAliasChecker extends ContextVisitor
{
	private SJTypeSystem sjts = (SJTypeSystem) typeSystem();
	//private SJNodeFactory sjnf = (SJNodeFactory) nodeFactory();

	private SJContext sjContext = new SJContext_c(sjts);

	public SJAliasChecker(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);
	}

	protected NodeVisitor enterCall(Node parent, Node n) throws SemanticException
	{
		if (n instanceof ClassDecl) // Duplicated from SJContextVisitor.
		{
			sjContext.pushClass((ClassDecl) n);
		}
		else if (n instanceof ProcedureDecl) // Should make more fine-grained, e.g. try-scopes, branch cases, etc.
		{
			sjContext.pushCode();
		}

		return this;
	}

	protected Node leaveCall(Node parent, Node old, Node n, NodeVisitor v) throws SemanticException
	{
		if (n instanceof ClassDecl || n instanceof ProcedureDecl)
		{
			sjContext.pop();
		}
		else if (n instanceof FieldDecl)
		{
			n = checkFieldDecl((FieldDecl) n);
		}
		else if (n instanceof LocalDecl)
		{
			n = checkLocalDecl((LocalDecl) n);
		}
		else if (n instanceof Assign)
		{
			n = checkAssign((Assign) n);
		}
		else if (n instanceof Cast)
		{
			n = checkCast((Cast) n);
		}

		return n;
	}

	private FieldDecl checkFieldDecl(FieldDecl fd) throws SemanticException
	{ // Uninitialised session field variables prevented by SJCreateOperationParser.
		Expr init = fd.init();

		if (init != null)
		{
			Type initType = init.type();

			if (!initType.isNull())
			{
				if (initType.typeEquals(SJConstants.SJCHANNEL_TYPE))
				{
					fd = (FieldDecl) checkSJChannelDecl(fd, fd.declType(), init, fd.name());
				}
				else if (initType.isSubtype(SJConstants.SJSOCKET_TYPE) || initType.isSubtype(SJConstants.SJSERVER_TYPE))
				{
					throw new SemanticException("Session socket/server field variables not allowed: " + init);
				}
			}
		}

		return fd;
	}

	private LocalDecl checkLocalDecl(LocalDecl ld) throws SemanticException
	{
		Expr init = ld.init();
		if (init != null)
		{
			Type initType = init.type();

			if (!initType.isNull())
			{
				if(initType.isSubtype(SJConstants.SJCHANNEL_TYPE))
				{
					ld = (LocalDecl) checkSJChannelDecl(ld, ld.type().type(), init, ld.name());
				}
				else if(initType.isSubtype(SJConstants.SJSOCKET_TYPE))
				{
					ld = (LocalDecl) checkSJSocketDecl(ld, ld.type().type(), init, ld.name());
				}
				else if(initType.isSubtype(SJConstants.SJSERVER_TYPE))
				{
					ld = (LocalDecl) checkSJServerDecl(ld, ld.type().type(), init, ld.name());
				}
			}
		}

		return ld;
	}

	private Node checkSJChannelDecl(Node n, Type declType, Expr init, String name) throws SemanticException // Although it is a declaration from session typing perspective, in terms of actual Java it can also be an assign to an uninitialised (null) variable. // Node n only used for feedback.
	{
		if (!declType.typeEquals(SJConstants.SJCHANNEL_TYPE))
		{
			throw new SemanticException("Bad target type: " + declType);
		}

		if (sjContext.currentContext().hasChannel(name)) // This check only needed for assigns, implicitly true for decls by base typing.
		{
			throw new SemanticException("Cannot alias a server-address variable: " + n);
		}

		SJNamedInstance sjni = null;

		if (init instanceof SJChannelCreate)
		{
			sjni = SJCompilerUtils.getSJNamedInstance((SJChannelCreate) init);
		}
		else if (init instanceof SJReceiveChannel)
		{
			sjni = SJCompilerUtils.getSJReturnOpInstance((SJReceiveChannel) init);
		}
		else
		{
			throw new SemanticException("Cannot alias a server-address variable: " + n);
		}

		sjContext.addChannel(sjni);

		return n;
	}

	private Node checkSJSocketDecl(Node n, Type declType, Expr init, String name) throws SemanticException
	{
		if (!declType.isSubtype(SJConstants.SJSOCKET_TYPE)) // Doesn't need to be typeEquals?
		{
			throw new SemanticException("Bad target type: " + declType);
		}

		if (sjContext.currentContext().hasSocket(name))
		{
			throw new SemanticException("Cannot alias a session socket variable: " + n);
		}

		SJNamedInstance sjni = null;

		if (init instanceof SJSocketCreate)
		{
			sjni = SJCompilerUtils.getSJNamedInstance((SJSocketCreate) init);
		}
		else if (init instanceof SJAccept)
		{
			sjni = SJCompilerUtils.getSJReturnOpInstance((SJAccept) init);
		}
		else if (init instanceof SJReceiveSession)
		{
			sjni = SJCompilerUtils.getSJReturnOpInstance((SJReceiveSession) init);
		}
		else if (init instanceof SJCall)
		{
			SJCall sjc = (SJCall) init;

			//if (SJCompilerUtils.anonymousSJReturn(sjc) // Should also be done in above cases.
			{
				//SJMethodInstance sjmi = SJCompilerUtils.getSJCallMethodInstance(sjc);

				sjni = sjts.SJNamedInstance(sjc.position(), SJCompilerUtils.getSJCallReturnType(sjc), SJCompilerUtils.getSJCallReturnName(sjc));
			}
		}
		else
		{
			throw new SemanticException("Cannot alias a session socket variable: " + n);
		}

		sjContext.addSocket(sjni);

		return n;
	}

	private Node checkSJServerDecl(Node n, Type declType, Expr init, String name) throws SemanticException
	{
		if (!declType.isSubtype(SJConstants.SJSERVER_TYPE))
		{
			throw new SemanticException("Bad target type: " + declType);
		}

		if (sjContext.currentContext().hasSocket(name))
		{
			throw new SemanticException("Cannot alias a session server variable: " + n);
		}

		SJNamedInstance sjni = null;

		if (init instanceof SJServerCreate)
		{
			sjni = SJCompilerUtils.getSJNamedInstance((SJServerCreate) init);
		}
		else
		{
			throw new SemanticException("Cannot alias a server-address variable: " + n);
		}

		sjContext.addServer(sjni);

		return n;
	}

	private Assign checkAssign(Assign assign) throws SemanticException
	{
		Expr left = assign.left();
		Expr right = assign.right(); // Can't be null.

		if (left.type().isSubtype(SJConstants.SJCHANNEL_TYPE)) // Actually, assignment of channel-create prevented by SJCreateOperationParser.
		{
			checkSJChannelDecl(assign, assign.type(), right, ((NamedVariable) left).name());
		}
		else if (left.type().isSubtype(SJConstants.SJSOCKET_TYPE))
		{
			checkSJSocketDecl(assign, assign.type(), right, ((NamedVariable) left).name()); // Allows "assignment-initialisation" of socket variables originally declared as null.
		}
		else if (left.type().isSubtype(SJConstants.SJSERVER_TYPE))
		{
			checkSJServerDecl(assign, assign.type(), right, ((NamedVariable) left).name());
		}
		else
		{
			Type rightType = right.type();

			if (!rightType.isNull())
			{
				if (rightType.isSubtype(SJConstants.SJCHANNEL_TYPE))
				{
					throw new SemanticException("Cannot alias a server-address variable: " +
					assign);
				}
				else if (rightType.isSubtype(SJConstants.SJSOCKET_TYPE))
				{
					throw new SemanticException("Cannot alias a session socket variable: " +
					assign);
				}
				else if (rightType.isSubtype(SJConstants.SJSERVER_TYPE))
				{
					throw new SemanticException("Cannot alias a session server variable: " +
					assign);
				}
			}
		}

		return assign;
	}

	private Cast checkCast(Cast cast) throws SemanticException
	{
		if (cast.expr().type().isSubtype(SJConstants.SJCHANNEL_TYPE)) // SJCast is dropped by preceding passes (session type building).
		{
			throw new SemanticException("Cannot cast a server address variable: " + cast);
		}
		else if (cast.expr().type().isSubtype(SJConstants.SJSOCKET_TYPE))
		{
			throw new SemanticException("Cannot cast a session socket variable: " + cast);
		}
		else if (cast.expr().type().isSubtype(SJConstants.SJSERVER_TYPE))
		{
			throw new SemanticException("Cannot cast a session server variable: " + cast);
		}

		return cast;
	}
}
