/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import sj.SJConstants;
import sj.ast.*;
import sj.ast.protocoldecls.*;
import sj.ast.createops.*;
import sj.ast.sessops.structuralops.*;
import sj.ast.proceduredecls.*;
import sj.types.*;
import sj.types.typeobjects.sesstypes.*;
import sj.types.typeobjects.typeableinstances.*;
import sj.visit.contexts.*;

import sj.util.SJCompilerUtils;

// SJTypeBuilder shouldn't be a SJContextVisitor. This class should only be used to help SJTypeChecker. // Maybe it is ok - this class is responsible for automatically managing and recording static context information, i.e. channels, sockets and servers. SJTypeChecker extends it with the routines needed for the dynamic information, i.e. session implementations.
public abstract class SJContextVisitor extends ContextVisitor
{
	private SJTypeSystem sjts = typeSystem();
	//private SJNodeFactory sjnf = nodeFactory();

	private SJContext sjc = new SJContext_c(sjts);

	public SJContextVisitor(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);
	}

	public SJTypeSystem typeSystem()
	{
		return (SJTypeSystem) super.typeSystem();
	}

	public SJNodeFactory nodeFactory()
	{
		return (SJNodeFactory) super.nodeFactory();
	}

	public SJContext sjContext()
	{
		return sjc;
	}

	protected final NodeVisitor enterCall(Node parent, Node n) throws SemanticException
	{
		if (n instanceof ClassDecl) // Need to record session type information for classes and methods (after it has been built).
		{
			enterClassScope((ClassDecl) n);
		}
		else if (n instanceof New)
		{
			New newn = (New) n;

			if (newn.body() != null) // Safe to allow (maybe, given current access restrictions).
			{
				if (newn.type().isSubtype(SJConstants.SJTHREAD_TYPE))
				{
					throw new SemanticException("Anonymous SJThreads not done yet: " + newn);
				}

				System.out.println("Warning: anonymous classes not finished yet: " + newn);

				//enterClassScope(...); // Should be particularly useful for SJThreads.
			}
		}
		else if (n instanceof ProcedureDecl)
		{
			if (n instanceof SJProcedureDecl)
			{
				/*if (n instanceof SJConstructorDecl)
				{

				}
				else*/ if (n instanceof SJMethodDecl)
				{
					enterSJMethodScope((SJMethodDecl) n);
				}
			}
			else
			{
				sjc.pushCode();
			}
		}
		else if (n instanceof Try) // !(this instanceof SJTypeChecker) // SJContextVisitor doesn't maintain contexts for session-try and session operations. But standard scope rules prevent the use of session variables from the wrong scope, like a subsequent session-try or in/outwhile - relying on exisiting entries to be overwritten in such cases.
		{
			if (n instanceof SJSessionTry)
			{
				enterSJSessionTryScope((SJSessionTry) n);
			}
			else
			{
				sjc.pushTry();
			}
		}
		/*else if (n instanceof If) // Needs overriding to push and pop branch case contexts.
		{

		}*/
		else if (n instanceof While && !(n instanceof SJInwhile || n instanceof SJOutwhile))
		{
			enterWhileScope((While) n);
		}
		else if (n instanceof SJStructuralOperation) // Session operations need specific operation type information - this is only tracked by the actual session type checker.
		{
			enterSJStructuralScope((SJStructuralOperation) n);
		}

		return sjEnterCall(parent, n);
	}

	abstract protected SJContextVisitor sjEnterCall(Node parent, Node n) throws SemanticException;

	protected void enterClassScope(ClassDecl cd) throws SemanticException
	{
		sjc.pushClass(cd); // Session type information is not built for classes (no session-class concept).
	}

	protected void enterSJMethodScope(SJMethodDecl sjmd)
	{
		SJMethodInstance sjmi = SJCompilerUtils.getSJMethodInstance(sjmd);

		if (sjmi == null)
		{
			sjc.pushCode();
		}
		else
		{
			if (sjmd instanceof SJRecursiveMethodDecl)
			{
				sjc.pushSJRecursiveMethodDecl(sjmi);
			}
			else
			{
				sjc.pushSJMethodDecl(sjmi);
			}
		}
	}

	protected void enterSJSessionTryScope(SJSessionTry sjst) throws SemanticException // To be overridden if required.
	{
		 // By default, no context pushed (it's a "dynamic" scope which can be nested within other session scopes - so we need to update the outer context when we leave and pop this one, which does not fit into the current SJContextVisitor pattern).
	}

	protected void enterWhileScope(While whilen) throws SemanticException
	{
		sjc.pushWhile(); // Scope pushed for non-session contexts. // Actually While is a session scope, but it is "isolated" because it cannot affect outer scopes due to typing rules.
	}

	protected void enterSJStructuralScope(SJStructuralOperation sjso) throws SemanticException
	{
		// Dynamic contexts, like session-try scopes.
	}

	protected final Node leaveCall(Node parent, Node old, Node n, NodeVisitor v) throws SemanticException
	{
		Node visited = sjLeaveCall(parent, old, n, v); // Need a convention that contexts not manipulated in here (unless paired up with extra work when scope entered), because should be done by overriding the leave scope methods. Otherwise a further subclass might do so but would then get unexpected (inherited) behaviour. // First do stuff within the context of the current scope...

		// ...and then pop the context to leave the scope.
		if (visited instanceof ClassDecl)
		{
			 visited = leaveClassScope((ClassDecl) visited); // Should generalise to allow anonymous classes.
		}
		/*else if (visited instanceof New)
		{
			For anonymous classes.
		}*/
		else if (visited instanceof ProcedureDecl)
		{
			if (visited instanceof SJProcedureDecl)
			{
				/*if (visited instanceof SJConstructorDecl)
				{

				}
				else*/ if (visited instanceof SJMethodDecl)
				{
					visited = leaveSJMethodScope((SJMethodDecl) visited);
				}
			}
			else
			{
				sjc.pop();
			}
		}
		else if (visited instanceof Try) // !(this instanceof SJTypeChecker)
		{
			if (visited instanceof SJSessionTry)
			{
				visited = leaveSJSessionTryScope((SJSessionTry) visited);
			}
			else
			{
				sjc.pop(); // Problem: the popped context isn't immediately accessible to visitors, so contexts like session-try and structural operations, where the just-visited context may be needed to update the outer context, become tricky to visit. So that's why these contexts are not directly handled by SJContextVisitor, but it supports enter/exit methods that can be overidden in subclasses. Basically, these for these contexts we need to "do something, pop context, do something" whereas the basic visitor structure of SJContextVisitor is only "do something, pop context". Can encode by having special pop routines that include the final "do something" (like popBranchCaseContext), but that's not clean.
			}
		}
		else if (visited instanceof While && !(n instanceof SJInwhile || n instanceof SJOutwhile))
		{
			visited = leaveWhileScope((While) n);
		}
		else if (visited instanceof SJStructuralOperation)
		{
			visited = leaveSJStructuralScope((SJStructuralOperation) visited);
		}
		else if (visited instanceof SJNamed)
		{
			visited = recordSJNamed((SJNamed) visited);
		}

		return visited;
	}

	abstract protected Node sjLeaveCall(Node parent, Node old, Node n, NodeVisitor v) throws SemanticException;

	protected ClassDecl leaveClassScope(ClassDecl cd)
	{
		sjc.pop();

		return cd;
	}

	protected SJMethodDecl leaveSJMethodScope(SJMethodDecl sjmd)
	{
		sjc.pop();

		return sjmd;
	}

	protected SJSessionTry leaveSJSessionTryScope(SJSessionTry sjst) throws SemanticException // To be overridden if required.
	{
		return sjst; // By default, no context pushed when scope entered.
	}

	protected While leaveWhileScope(While whilen) throws SemanticException
	{
		sjc.pop();

		return whilen;
	}

	protected SJStructuralOperation leaveSJStructuralScope(SJStructuralOperation sjso) throws SemanticException
	{
		return sjso;
	}

	private SJNamed recordSJNamed(SJNamed n) throws SemanticException
	{
		SJNamedInstance sjni = SJCompilerUtils.getSJNamedInstance((SJNamed) n);

		if (sjni == null) // Should we have this check? Or only use SJContextVisitor after session type information has been built?
		{
			return n; // Session type information not yet built.
		}

		if (n instanceof SJProtocolDecl) // The "static" session type information.
		{
			sjc.addProtocol(sjni);
		}
		else if (n instanceof SJChannelCreate)
		{
			sjc.addChannel(sjni);
		}
		else if (n instanceof SJSocketCreate)
		{
			if (sjni.sjType() instanceof SJUnknownType) // Socket was created using a received channel. So type only known in and after SJTypeChecker pass.
			{
				SJSocketCreate sjsc = (SJSocketCreate) n;

				try // Hacky? If the channel-receive is to be inferred, then the channel type is still unknown until SJTypeChecker.
				{
					sjni = sjni.sjType(sjc.channelType(sjsc.arguments().get(0).toString()));

					n = SJCompilerUtils.setSJNamedInstance(sjsc, sjni);
				}
				catch (SemanticException se)
				{

				}
			}

			sjc.addSocket(sjni);
		}
		else if (n instanceof SJServerCreate)
		{
			sjc.addServer(sjni);
		}
		else
		{
			// Session variables., server operations, channel/session return, ...
		}

		return n;
	}

	public Node override(Node parent, Node n)
	{
		try
		{
			if (n instanceof If)
			{
				return overrideIf((If) n);
			}
		}
		catch (SemanticException se)
		{
			enqueueError(n, se); // Should stop all further compilation? Using bypass?
		}

		return null;
	}

	protected Node overrideIf(If ifn) throws SemanticException // To be overridden if required. // Probably better to "override" (using a del object) the visitChildren method of If to store the results from each case somewhere.
	{
		ifn = ifn.cond((Expr) ifn.cond().visit(this));

		sjc.pushIf();
		sjc.pushIfBranch();

		Stmt consequent = ifn.consequent();

		if (consequent != null)
		{
			ifn = ifn.consequent((Stmt) consequent.visit(this));
		}

		sjc.popBranchCaseContext();
		sjc.pushIfBranch();

		Stmt alternative = ifn.alternative();

		if (alternative != null)
		{
			ifn = ifn.alternative((Stmt) alternative.visit(this));
		}

		sjc.popBranchCaseContext();
		sjc.pop();

		return ifn;
	}

	// Needed because 'override' can't throw (Semantic)Exception.
	protected void enqueueError(Node n, SemanticException se)
	{
		if (se.getMessage() != null)
		{
			Position position = se.position();

			if (position == null)
			{
				position = n.position();
			}

			errorQueue().enqueue(ErrorInfo.SEMANTIC_ERROR, se.getMessage(), position);
		}

		this.error = catchErrors(n) ? false : true;
	}
}
