/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import java.util.*;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.Position;
import polyglot.visit.ContextVisitor;
import polyglot.visit.NodeVisitor;

import sj.SJConstants;
import sj.ast.*;
import sj.ast.createops.*;
import sj.util.*;

// Need to be careful for nested create operations? e.g. a server-address into a socket create. // Also prevents session socket/server field declarations and assignment of channel-create operations.
public class SJCreateOperationParser extends ContextVisitor
{
	private SJNodeFactory sjnf = (SJNodeFactory) nodeFactory();

	private SJTypeBuilder sjtb = new SJTypeBuilder(job(), typeSystem(), nodeFactory());

	public SJCreateOperationParser(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);
	}

	protected Node leaveCall(Node old, Node n, NodeVisitor v) throws SemanticException
	{
		if (n instanceof Call)
		{
			Call call = (Call) n;

			if (call.name().equals(SJConstants.SJ_SYNTAX_CREATE))
			{
				n = parseCreateCall(call);
			}
		}
		else if (n instanceof VarInit)
		{
			n = (Node) visitVarInit((VarInit) n);
		}
		else if (n instanceof Assign)
		{
			n = visitAssign((Assign) n);
		}

		return n;
	}

	private Node parseCreateCall(Call call) throws SemanticException
	{
		Receiver target = call.target();
		Type targetType = target.type();
		Position pos = call.position();

		Node sjn = null;

		if (targetType.typeEquals(SJConstants.SJCHANNEL_TYPE))
		{
			sjn = sjnf.SJChannelCreate(pos, target, (List<Expr>) call.arguments());
		}
		else if (targetType.isSubtype(SJConstants.SJSOCKET_TYPE))
		{
			sjn = sjnf.SJSocketCreate(pos, target, (List<Expr>) call.arguments());
		}
		else if (targetType.isSubtype(SJConstants.SJSERVER_TYPE))
		{
			sjn = sjnf.SJServerCreate(pos, target, (List<Expr>) call.arguments());
		}

		if (sjn != null)
		{
			sjn = sjtb.buildTypes(context(), sjn);

			SJCompilerUtils.debugPrintln("Parsed: " + sjn );

			return sjn;
		}
		else
		{
			return call;
		}
	}

	private VarInit visitVarInit(VarInit vi) throws SemanticException
	{
		if (vi instanceof FieldDecl)
		{
			FieldDecl fd = (FieldDecl) vi;
			Type type = fd.declType();
			Expr init = fd.init();

			if (type.typeEquals(SJConstants.SJCHANNEL_TYPE) || type.isSubtype(SJConstants.SJSOCKET_TYPE) || type.isSubtype(SJConstants.SJSERVER_TYPE)) // Move to alias checker or ...? // SJProtocolDecl flags checked by type builder.
			{
				if (!fd.flags().isPrivate())
				{
					throw new SemanticException("Non-private session field variables not done yet: " + vi);
				}

				if (init == null)
				{
					throw new SemanticException("Uninitialised session field variables not allowed: " + fd);
				}
			}

			if (init instanceof SJCreateOperation)
			{
				//Position pos = fd.position();
				String name = fd.name();

				if (init instanceof SJChannelCreate)
				{
					fd = fd.init((SJChannelCreate) SJCompilerUtils.setSJNamedName((SJChannelCreate) init, name));
				}
				else if (init instanceof SJSocketCreate || init instanceof SJServerCreate)
				{
					throw new SemanticException("Field session sockets/servers not allowed: " + vi);
				}

				//vi = fd.flags(fd.flags().Final()); // Move to a translation pass? // Variable may be overwritten with null reference by SJHigherOrderTranslator.
				vi = fd;
			}
		}
		else //if (vi instanceof LocalDecl)
		{
			LocalDecl ld = (LocalDecl) vi;
			Expr init = ld.init();

			if (init instanceof SJCreateOperation)
			{
				ld = ld.init((SJCreateOperation) SJCompilerUtils.setSJNamedName((SJCreateOperation) init, ld.name())); // All create operatioms permitted in local context.

				//vi = ld.flags(ld.flags().Final());
				vi = ld;
			}
		}

		return vi;
	}

	private Assign visitAssign(Assign ass) throws SemanticException
	{
		Expr right = ass.right();

		if (right instanceof SJChannelCreate) // General (re-)assignment to field variables prevented by alias checker.
		{
			throw new SemanticException("Assignment of channel-create not permitted: " + ass);
		}
		else if (right instanceof SJSocketCreate || right instanceof SJServerCreate)
		{
			String name = ((NamedVariable) ass.left()).name();

			ass = ass.right((SJCreateOperation) SJCompilerUtils.setSJNamedName((SJCreateOperation)right, name));
		}

		return ass;
	}
}
