/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import java.util.*;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import sj.SJConstants;
import sj.ast.*;
import sj.ast.proceduredecls.*;
import sj.ast.typenodes.*;
import sj.ast.sessformals.*;
import sj.ast.socketops.*;
import sj.types.SJTypeSystem;
import sj.util.*;

// Must come before SJThreadTranslator or else run method formal types won't get serialized.
public class SJProcedureDeclTranslator extends ErrorHandlingVisitor
{
	private SJTypeSystem sjts = (SJTypeSystem) typeSystem();
	private SJNodeFactory sjnf = (SJNodeFactory) nodeFactory();

	private SJTypeEncoder sjte = new SJTypeEncoder(sjts);

	private Type stringType;

	public SJProcedureDeclTranslator(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);

		try
		{
			stringType = sjts.typeForName(SJConstants.JAVA_STRING_CLASS);
		}
		catch (SemanticException se)
		{
			throw new RuntimeException(se);
		}
	}

	protected NodeVisitor enterCall(Node parent, Node n) throws SemanticException
	{
		return this;
	}

	protected Node leaveCall(Node old, Node n, NodeVisitor v) throws SemanticException
	{
		if (n instanceof ClassDecl)
		{
			n = translateClassDecl((ClassDecl) n); // Pops serializedFormals.
		}

		return n;
	}

	private ClassDecl translateClassDecl(ClassDecl cd) throws SemanticException // Should extend to support anonymous classes too.
	{
		ClassBody body = cd.body();

		List<ClassMember> members = new ArrayList<ClassMember>();

		//members.addAll(body.members());

		for (Object member : body.members())
		{
			if (member instanceof SJProcedureDecl)
			{
				/*if (member instanceof SJConstructorDecl)
				{

				}
				else*/ if (member instanceof SJMethodDecl)
				{
					SJMethodDecl sjmd = (SJMethodDecl) member;
					String mname = sjmd.name();
					Flags flags = sjmd.flags();

					if (!flags.isFinal())
					{
						flags = flags.Final();
					}

					//SJMethodInstance sjmi = SJCompilerUtils.getSJMethodInstance(member); // Doesn't have the formal names, only their types.

					List<Stmt> sockops = new LinkedList<Stmt>();

					for (SJFormal sjFormal : SJCompilerUtils.getSJMethodDeclFormals(sjmd))
					{
						//if (sjFormal instanceof SJTypeFormal)
						{
							SJTypeFormal sjtf = (SJTypeFormal) sjFormal;
							Position pos = sjtf.position();

							String fname = SJConstants.SJ_FORMAL_PREFIX + sjmd.name() + "_" + sjtf.name();
							String encoded = sjte.encode(SJCompilerUtils.getSJType(sjtf));
							TypeNode stringTypeNode = sjnf.CanonicalTypeNode(pos, stringType);

							members.add(sjnf.FieldDecl(pos, flags, stringTypeNode, sjnf.Id(pos, fname), sjnf.StringLit(pos, encoded)));

							if (sjmd instanceof SJRecursiveMethodDecl) // Convenient to do this here, but maybe would belong more logically to SJHigherOrderTranslator, or its own pass. // The counterpart to this translation routine (for the actual recursive session method calls) is done in SJHigherOrderTranslator.
							{
								SJSocketOperation sockop = sjnf.SJRecursionFalse(pos, sjnf.SJLocalSocket(pos, sjnf.Id(pos, sjtf.name())), new SJLabel(mname));

								sockops.add(sjnf.Eval(pos, sockop));
							}
						}
					}

					SJTypeNode sjReturnType = SJCompilerUtils.getSJMethodDeclReturnType(sjmd);

					if (sjReturnType != null)
					{
						Position pos = sjReturnType.position();
						String fname = SJConstants.SJ_RETURN_TYPE_PREFIX + sjmd.name();
						String encoded = sjte.encode(sjReturnType.type());
						TypeNode stringTypeNode = sjnf.CanonicalTypeNode(pos, stringType);

						members.add(sjnf.FieldDecl(pos, flags, stringTypeNode, sjnf.Id(pos, fname), sjnf.StringLit(pos, encoded)));
					}

					if (sjmd instanceof SJRecursiveMethodDecl)
					{
						sockops.addAll((List<Stmt>) sjmd.body().statements());

						members.add((SJMethodDecl) ((SJMethodDecl) sjmd.body(sjnf.Block(sjmd.body().position(), sockops))).flags(flags));
					}
					else
					{
						members.add(sjmd);
					}
				}
				else
				{
					members.add((ClassMember) member);
				}
			}
			else // Currently needed because constructors not done yet.
			{
				members.add((ClassMember) member);
			}
		}

		return cd.body(body.members(members));
	}
}
