/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import java.util.*;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.ContextVisitor;
import polyglot.visit.NodeVisitor;

import sj.SJConstants;
import sj.ast.*;
import sj.ast.protocoldecls.*;
import sj.types.SJTypeSystem;
import sj.types.typeobjects.typeableinstances.SJNamedInstance;
import sj.util.*;

public class SJProtocolDeclTranslator extends ContextVisitor // Subsequent ContextVisitor passes are broken because type building not done for new nodes? // Recording a reference to the containing AST node inside typeable instance objects might be useful for translation, maybe no need to manually track context scopes.
{
	private SJTypeSystem sjts = (SJTypeSystem) typeSystem();
	private SJNodeFactory sjnf = (SJNodeFactory) nodeFactory();

	//private SJTypeBuilder sjtb = new SJTypeBuilder(job(), typeSystem(), nodeFactory());

	private SJTypeEncoder sjte = new SJTypeEncoder(sjts);

	private Stack<HashMap<Id, Term>> protocolDeclContexts = new Stack<HashMap<Id, Term>>(); // For both field (class) and local scopes.

	private Type stringType; // Move to SJConstants?

	public SJProtocolDeclTranslator(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);

		try
		{
			stringType = sjts.typeForName(SJConstants.JAVA_STRING_CLASS);
		}
		catch (SemanticException se)
		{
			throw new RuntimeException(se);
		}
	}

	protected NodeVisitor enterCall(Node parent, Node n) throws SemanticException
	{
		if (n instanceof ClassBody)
		{
			protocolDeclContexts.push(new HashMap<Id, Term>());
		}
		else if (n instanceof ProcedureDecl)
		{
			protocolDeclContexts.push(new HashMap<Id, Term>());
		}

		return this;
	}

	protected Node leaveCall(Node old, Node n, NodeVisitor v) throws SemanticException
	{
		if (n instanceof SJProtocolDecl)
		{
			n = visitSJProtocolDecl((SJProtocolDecl) n);
		}
		//else if (n instanceof ClassDecl)
		else if (n instanceof ClassBody) // Supports anonymous inner classes.
		{
			n = visitClassBody((ClassBody) n);
		}
		else if (n instanceof ProcedureDecl)
		{
			n = visitProcedureDecl((ProcedureDecl) n);
		}

		return n;
	}

	private SJProtocolDecl visitSJProtocolDecl(SJProtocolDecl pd) throws SemanticException
	{
		Position pos = pd.position();
		Flags flags = null;

		SJNamedInstance sjni = SJCompilerUtils.getSJNamedInstance(pd);

		Id pid = sjnf.Id(pos, SJConstants.SJ_STRING_DECL_PREFIX + sjni.name());
		TypeNode stringTypeNode = sjnf.CanonicalTypeNode(pos, stringType);
		StringLit encodedSJType = sjnf.StringLit(pos, sjte.encode(sjni.sjType()));

		Term protocolDecl = null;

		if (pd instanceof SJFieldProtocolDecl)
		{
			flags = ((SJFieldProtocolDecl) pd).flags(); // Already set to final by node factory.

			protocolDecl = sjnf.FieldDecl(pos, flags, stringTypeNode, pid, encodedSJType);
		}
		else if (pd instanceof SJLocalProtocolDecl)
		{
			flags = ((SJLocalProtocolDecl) pd).flags();

			protocolDecl = sjnf.LocalDecl(pos, flags, stringTypeNode, pid, encodedSJType);
		}

		//protocolDecl = (Term) sjtb.buildTypes(context(), protocolDecl); // Broken for LocalDecls due to modification of context state by the compilation passes of SJTypeBuilder.

		protocolDeclContexts.peek().put(pid, protocolDecl);

		TypeNode sjProtocolTypeNode = sjnf.CanonicalTypeNode(pos, SJConstants.SJPROTOCOL_TYPE);

		List<Expr> args = new LinkedList<Expr>();

		if (pd instanceof SJFieldProtocolDecl)
		{
			Receiver target = null;

			if (flags.isStatic())
			{
				target = sjnf.CanonicalTypeNode(pos, context().currentClass());
			}
			else
			{
				target = sjnf.Special(pos, Special.THIS);
			}

			args.add(sjnf.Field(pos, target, pid));

			pd = ((SJFieldProtocolDecl) pd).init(sjnf.New(pos, sjProtocolTypeNode, args));
		}
		else if (pd instanceof SJLocalProtocolDecl)
		{
			args.add(sjnf.Local(pos, pid));

			pd = ((SJLocalProtocolDecl) pd).init(sjnf.New(pos, sjProtocolTypeNode, args));
		}

		return pd;
		//return (SJProtocolDecl) sjtb.buildTypes(context(), pd); // Should do type building for future ContextVisitor passes? // But problems building types when referencing variables that are have just been inserted (in the same pass).
	}

	//private ClassDecl visitClassBody(ClassDecl cd)
	private ClassBody visitClassBody(ClassBody body)
	{
		HashMap<Id, Term> protocols = (HashMap<Id, Term>) protocolDeclContexts.pop();

		if (protocols.isEmpty())
		{
			//return cd;
			return body;
		}

		//ClassBody body = cd.body();

		List<ClassMember> classMembers = new LinkedList<ClassMember>();

		for (Id pid : protocols.keySet())
		{
			classMembers.add((FieldDecl) protocols.get(pid));
		}

		classMembers.addAll((List<ClassMember>) body.members());

		//return (ClassDecl) cd.body(body.members(classMembers));
		return body.members(classMembers);
	}

	private ProcedureDecl visitProcedureDecl(ProcedureDecl pd)
	{
		HashMap<Id, Term> protocols = (HashMap<Id, Term>) protocolDeclContexts.pop();

		if (pd instanceof MethodDecl)
		{
			for (Id pid : protocols.keySet())
			{
				pd = (ProcedureDecl) pd.body(pd.body().prepend((Stmt) protocols.get(pid)));
			}
		}
		else //if (pd instanceof ConstructorDecl)
		{
			List<Stmt> statements = new LinkedList<Stmt>();

			Block body = pd.body();

			statements.add((Stmt) body.statements().get(0)); // 'super' call.

			for (Id pid : protocols.keySet())
			{
				statements.add((Stmt) protocols.get(pid));
			}

			for (int i = 1; i < body.statements().size(); i++)
			{
				statements.add((Stmt) body.statements().get(i));
			}

			pd = (ProcedureDecl) pd.body(body.statements(statements));
		}

		return pd;
	}
}
