/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import java.util.*;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.Position;
import polyglot.visit.ContextVisitor;
import polyglot.visit.NodeVisitor;

import sj.SJConstants;
import sj.ast.*;
import sj.ast.protocoldecls.*;
import sj.ast.typenodes.*;
import sj.ast.createops.*;
import sj.ast.sessvars.*;
import sj.ast.sessops.*;
import sj.ast.sessops.basicops.*;
import sj.ast.sessops.structuralops.*;
import sj.ast.servops.*;
import sj.ast.sesscasts.*;
import sj.util.*;

/**
 * Also parses server operations, session-return operations, session method-calls and session-spawns.
 */
public class SJSessionOperationParser extends ContextVisitor
{
	//private SJTypeSystem sjts = (SJTypeSystem) typeSystem();
	private SJNodeFactory sjnf = (SJNodeFactory) nodeFactory();

	private SJTypeBuilder sjtb = new SJTypeBuilder(job(), typeSystem(), nodeFactory());

	private Stack<Set<String>> beginProtocols = new Stack<Set<String>>(); // HACK: to support e.g. (@p) s.receive(), where p is begin.... The node factory cannot tell what type p is so it makes a SJSocket cast rather than SJServerAddress. Can fix by recording seen (begin prefixed) protocol declarations (although their types are not properly built until SJTypeBuilder), checking what TypeNode p is and changing the cast type if necessary.

	public SJSessionOperationParser(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);
	}

	protected NodeVisitor enterCall(Node parent, Node n) throws SemanticException
	{
		if (n instanceof ClassDecl || n instanceof ProcedureDecl)
		{
			beginProtocols.push(new HashSet<String>());
		}

		return this;
	}

	protected Node leaveCall(Node parent, Node old, Node n, NodeVisitor v) throws SemanticException
	{
		if (n instanceof ClassDecl || n instanceof ProcedureDecl)
		{
			beginProtocols.pop();
		}
		else if (n instanceof SJProtocolDecl)
		{
			SJProtocolDecl sjpd = (SJProtocolDecl) n;

			if (SJCompilerUtils.getSJProtocolDeclTypeNode(sjpd) instanceof SJBeginNode)
			{
				beginProtocols.peek().add(SJCompilerUtils.getSJNamedName(sjpd));
			}
		}

		else if (n instanceof Call && !(n instanceof SJCreateOperation))
		{
			Call call = (Call) n;
			Type targetType = call.target().type();

			if (targetType.isSubtype(SJConstants.SJSOCKET_TYPE)) // Parsing basic session operations.
			{
				n = parseSJBasicOperation(parent, call); // Only basic operations are extended from Call.
			}
			else if (targetType.isSubtype(SJConstants.SJSERVER_TYPE))
			{
				n = parseSJServerOperation(parent, call);
			}
			//else if (targetType.isSubtype(SJConstants.SJTHREAD_TYPE)) // Breaks method calls inside a SJThread.
			else if (call instanceof SJSpawn)
			{
				n = parseSJSpawn((SJSpawn) call);
			}
			else
			{
				n = parseCall(parent, call); // May or may not be a SJCall.
			}
		}
		else if (n instanceof SJCast) // For type inference cases, receiveChannel/Session operation not parsed until session type checking.
		{
			n = parseSJCast(parent, (SJCast) n);
		}
		else if (n instanceof SJStructuralOperation)
		{
			n = parseSJStructuralOperation((SJStructuralOperation) n);
		}
		else if (n instanceof Return)
		{
			Return ret = (Return) n;
			Expr expr = ret.expr();

			if (expr != null && !(expr instanceof NullLit) && expr.type().isSubtype(SJConstants.SJSOCKET_TYPE)) // Maybe should also explicitly check for the other session entity types.
			{
				n = parseSJReturn(ret);
			}
		}

		return n;
	}

	private SJReturn parseSJReturn(Return ret) throws SemanticException
	{
		Expr expr = ret.expr();

		SJReturn sjr = null;

		if (expr instanceof SJCall)
		{
			sjr = sjnf.SJReturn(ret.position(), expr);
		}
		else
		{
			if (expr instanceof Local)
			{
				SJLocalSocket sv = sjnf.SJLocalSocket(expr.position(), ((Local) expr).id());

				sjr = (SJReturn) sjnf.SJReturn(ret.position(), sv);
				sjr = (SJReturn) SJCompilerUtils.setSJNamedName(sjr, sv.name());
			}
			else
			{
				throw new SemanticException("Only local session variables can be returned: " + ret);
			}
		}

		sjr = (SJReturn) sjtb.buildTypes(context(), sjr);

		SJCompilerUtils.debugPrintln("Parsed: " + sjr);

		return sjr;
	}

	private Call parseSJBasicOperation(Node parent, Call call) throws SemanticException
	{
		Position pos = call.position();
		Receiver target = call.target();
		String callName = call.name(); // Maybe should use Id instead.

		SJBasicOperation sjbo = null;
		SJSocketVariable sv = null;

		if (target instanceof Local)
		{
			sv = sjnf.SJLocalSocket(target.position(), ((Local) target).id());
		}
		else //if (target instanceof Field) // Shouldn't get here.
		{
			throw new SemanticException("Field sockets not allowed: " + call);
		}

		// Translation of basic operations done implicitly by implementation of basic operation nodes as Call subclasses and by specifying the socket call targets here.
		if (callName.equals(SJConstants.SJ_SYNTAX_REQUEST))
		{
			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_REQUEST);

			sjbo = sjnf.SJRequest(pos, sv, name, (List<Expr>) call.arguments());
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_SEND))
		{
			Expr arg = (Expr) call.arguments().get(0); // Factor out constant.
			Type messageType = (Type) arg.type();

			if (messageType.isSubtype(SJConstants.SJCHANNEL_TYPE))
			{
				//throw new SemanticException("Inference for channel-sends not done yet: " + call);

				Id name = sjnf.Id(pos, SJConstants.SJSOCKET_SENDCHANNEL);

				sjbo = sjnf.SJSendChannel(pos, sv, name, (List<Expr>) call.arguments());
			}
			else if (messageType.isSubtype(SJConstants.SJSOCKET_TYPE))
			{
				sjbo = parseSJSendSession(call, sv);
			}
			else if (messageType.isBoolean()) // No receive equivalent here - return type overloading not possible. // But casted primitive-receives are inferred in type building.
			{
				Id name = sjnf.Id(pos, SJConstants.SJSOCKET_SENDBOOLEAN);

				sjbo = sjnf.SJSendBoolean(pos, sv, name, (List<Expr>) call.arguments());
			}
			else if (messageType.isInt())
			{
				Id name = sjnf.Id(pos, SJConstants.SJSOCKET_SENDINT);

				sjbo = sjnf.SJSendInt(pos, sv, name, (List<Expr>) call.arguments());
			}
			else if (messageType.isByte())
			{
				Id name = sjnf.Id(pos, SJConstants.SJSOCKET_SENDBYTE);

				sjbo = sjnf.SJSendByte(pos, sv, name, (List<Expr>) call.arguments());
			}
			else
			{
				Id name = sjnf.Id(pos, SJConstants.SJSOCKET_SEND);

				sjbo = sjnf.SJSend(pos, sv, name, (List<Expr>) call.arguments()); // Includes (to be inferred) session-send for now.
			}
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_SENDBOOLEAN))
		{
			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_SENDBOOLEAN);

			sjbo = sjnf.SJSendBoolean(pos, sv, name, (List<Expr>) call.arguments());
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_SENDINT))
		{
			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_SENDINT);

			sjbo = sjnf.SJSendInt(pos, sv, name, (List<Expr>) call.arguments());
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_SENDBYTE))
		{
			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_SENDBYTE);

			sjbo = sjnf.SJSendByte(pos, sv, name, (List<Expr>) call.arguments());
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_SENDCHANNEL))
		{
			if (!(call.arguments().get(0) instanceof NamedVariable)) // Didn't bother to make SJChannelVariable and parse the arguments (not really necessary).
			{
				throw new SemanticException("Can only pass named SJServerAddress variables: " + sjbo);
			}

			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_SENDCHANNEL);

			sjbo = sjnf.SJSendChannel(pos, sv, name, (List<Expr>) call.arguments());
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_SENDSESSION))
		{
			sjbo = parseSJSendSession(call, sv);
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_RECEIVE))
		{
			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_RECEIVE);

			sjbo = sjnf.SJReceive(pos, sv, name, (List<Expr>) call.arguments()); // Includes (to be inferred) channel and session-receive for now.
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_RECEIVEBOOLEAN))
		{
			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_RECEIVEBOOLEAN);

			sjbo = sjnf.SJReceiveBoolean(pos, sv, name, (List<Expr>) call.arguments());
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_RECEIVEINT))
		{
			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_RECEIVEINT);

			sjbo = sjnf.SJReceiveInt(pos, sv, name, (List<Expr>) call.arguments());
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_RECEIVEBYTE))
		{
			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_RECEIVEBYTE);

			sjbo = sjnf.SJReceiveByte(pos, sv, name, (List<Expr>) call.arguments());
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_RECEIVESESSION))
		{
			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_RECEIVESESSION);
			SJReceiveSession sjrs = sjnf.SJReceiveSession(pos, sv, name, call.arguments());

			if (!(parent instanceof SJCast))
			{
				String returned = SJCompilerUtils.parseSJReturnName(parent);

				sjbo = (SJReceiveSession) SJCompilerUtils.setSJReturnName(sjrs, returned);
			}
			else
			{
				sjbo = sjrs;
			}
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_RECEIVECHANNEL))
		{
			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_RECEIVECHANNEL);
			SJReceiveChannel sjrc = sjnf.SJReceiveChannel(pos, sv, name, (List<Expr>) call.arguments());

			if (!(parent instanceof SJCast))
			{
				String returned = SJCompilerUtils.parseSJReturnName(parent);

				sjbo = (SJReceiveChannel) SJCompilerUtils.setSJReturnName(sjrc, returned);
			}
			else
			{
				sjbo = sjrc;
			}
		}
		else if (callName.equals(SJConstants.SJ_SYNTAX_RECURSE))
		{
			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_RECURSE);
			SJLabel lab = new SJLabel((String) ((Expr) call.arguments().get(0)).constantValue());

			sjbo = sjnf.SJRecurse(pos, sv, name, call.arguments(), lab); // The label is left encoded as an argument.
		}
		else
		{
			// Socket creation and parameter operations.
		}

		if (sjbo != null) // Is this check necessary?
		{
			call = (SJBasicOperation) sjtb.buildTypes(context(), sjbo);

			SJCompilerUtils.debugPrintln("Parsed: " + call);
		}

		return call;
	}

	private SJSendSession parseSJSendSession(Call call, SJSocketVariable sv) throws SemanticException
	{
		Position pos = call.position();

		Expr arg = (Expr) call.arguments().get(0); // Factor out constant.

		if (!(arg instanceof NamedVariable))
		{
			throw new SemanticException("Can only delegate named SJSocket variables: " + call);
		}

		Id name = sjnf.Id(pos, SJConstants.SJSOCKET_SENDSESSION);

		List<Expr> args = new ArrayList<Expr>();

		NamedVariable nv = (NamedVariable) arg;

		if (nv instanceof Local) // Do here, or at session type checking?
		{
			args.add(sjnf.SJLocalSocket(nv.position(), ((Local) nv).id()));
		}
		else if (nv instanceof Field)
		{
			throw new SemanticException("Field session sockets not allowed: " + call);
		}

		List arguments = call.arguments();

		args.addAll((List<Expr>) call.arguments().subList(1, arguments.size())); // Should be unnecessary.

		return sjnf.SJSendSession(pos, sv, name, args);
	}

	private SJCast parseSJCast(Node parent, SJCast sjc) throws SemanticException
	{
		Expr castExpr = sjc.expr();

		if (castExpr instanceof SJReceiveChannel || castExpr instanceof SJReceiveSession) // Needed because pass may be run multiple times over same nodes - see parseCall comment above. // Anyway, receiveChannel/Socket could still be explicitly cast by user.
		{
			SJReturnOperation sjro = (SJReturnOperation) castExpr;

			String recorded = SJCompilerUtils.getSJReturnName(sjro);
			String returned = SJCompilerUtils.parseSJReturnName(parent);

			if (recorded == null && returned != null) // Why would returned be null?
			{
				sjc = (SJCast) sjc.expr((SJBasicOperation) SJCompilerUtils.setSJReturnName(sjro, returned));
			}

			return sjc;
		}

		/*if (castExpr instanceof SJCall) // Do we actually need to do any parsing operation here?
		{
			throw new SemanticException("Session-cast for session-method calls not done yet: " + sjc);
		}
		else*/ if (!(castExpr instanceof SJReceive))
		{
			throw new SemanticException("Session-cast is for higher-order receive operations, not: " + sjc);
		}

		String returned = SJCompilerUtils.parseSJReturnName(parent);

		SJReceive sjr = (SJReceive) castExpr;
		Position pos = sjr.position();
		SJSocketVariable sv = (SJSocketVariable) sjr.target();

		SJBasicOperation sjbo = null;

		if (sjc.type().isSubtype(SJConstants.SJCHANNEL_TYPE)) // Actually, can never get in here: not possible at this stage for a channel type cast on a ordinary receive.
		{
			Id name = sjnf.Id(pos, SJConstants.SJSOCKET_RECEIVECHANNEL);

			sjbo = sjnf.SJReceiveChannel(pos, sv, name, (List<Expr>) sjr.arguments());
		}
		else //if (sjc.type().isSubtype(SJConstants.SJSOCKET_TYPE))
		{
			if (sjc instanceof SJTypeCast)
			{
				SJTypeNode castType = SJCompilerUtils.getSJTypeCastTypeNode((SJTypeCast) sjc);

				if (castType instanceof SJProtocolNode)
				{
					if (isBeginProtocol(((SJProtocolNode) castType).protocol()))
					{
						Id name = sjnf.Id(pos, SJConstants.SJSOCKET_RECEIVECHANNEL);

						sjbo = sjnf.SJReceiveChannel(pos, sv, name, (List<Expr>) sjr.arguments());

						sjc = (SJCast) sjc.castType(sjnf.CanonicalTypeNode(pos, SJConstants.SJCHANNEL_TYPE));
					}
				}
			}

			if (sjbo == null)
			{
				Id name = sjnf.Id(pos, SJConstants.SJSOCKET_RECEIVESESSION);

				sjbo = sjnf.SJReceiveSession(pos, sv, name, sjr.arguments());
			}
		}

		sjbo = (SJBasicOperation) SJCompilerUtils.setSJReturnName((SJReturnOperation) sjbo, returned);

		return (SJCast) sjc.expr((SJBasicOperation) sjtb.buildTypes(context(), sjbo)); // Keep the cast (to carry type info.); drop in type builder. // No need to rebuild types for the cast itself?
	}

	private boolean isBeginProtocol(String protocol)
	{
		for (Set<String> context : beginProtocols) // Searches contexts backwards.
		{
			if (context.contains(protocol))
			{
				return true;
			}
		}

		return false;
	}

	private SJStructuralOperation parseSJStructuralOperation(SJStructuralOperation sjso) throws SemanticException
	{
		Receiver target = SJCompilerUtils.getSJStructuralOpTarget(sjso);

		SJSocketVariable sv = null;

		if (target instanceof Local)
		{
			sv = sjnf.SJLocalSocket(target.position(), ((Local) target).id());
		}
		else //if (target instanceof Field)
		{
			throw new SemanticException("Field session sockets not allowed: " + sjso);
		}

		sv = (SJSocketVariable) sjtb.buildTypes(context(), sv);

		return SJCompilerUtils.setSJStructuralOpTarget(sjso, sv);
	}

	private Call parseSJServerOperation(Node parent, Call call) throws SemanticException
	{
		String callName = call.name();
		Receiver target = call.target();
		Position pos = call.position();

		SJServerOperation sjso = null;
		SJServerVariable sv = null;

		if (target instanceof Local)
		{
			sv = sjnf.SJLocalServer(target.position(), ((Local) target).id());
		}
		else
		{
			throw new SemanticException("Field servers not allowed: " + call); // Shouldn't get here.
		}

		if (callName.equals(SJConstants.SJ_SYNTAX_ACCEPT))
		{
			String returned = SJCompilerUtils.parseSJReturnName(parent);
			Id name = sjnf.Id(pos, SJConstants.SJSERVER_ACCEPT);
			SJAccept sja = sjnf.SJAccept(pos, sv, name, (List<Expr>) call.arguments());

			sjso = (SJAccept) SJCompilerUtils.setSJReturnName(sja, returned);
		}
		else
		{
			// Server creation and parameter operations.
		}

		if (sjso != null)
		{
			call = (SJServerOperation) sjtb.buildTypes(context(), sjso);

			SJCompilerUtils.debugPrintln("Parsed: " + call);
		}

		return call;
	}

	private Call parseCall(Node parent, Call call) throws SemanticException
	{
		if (call instanceof SJCall) // The isSubtype call (for class types) below some how causes this pass to be rerun (for calls to methods in the current class context). Why?? Because need to rerun pass for method checking (forward declarations)? But this is just call-argument checking? Similar check to that in parseSJCast. // Not needed now?
		{
			return call;
		}

		List args = call.arguments();
		List<Expr> sjargs = new LinkedList<Expr>();

		boolean isSJCall = false;

		for (int i = 0; i < args.size(); i++)
		{
			Expr arg = (Expr) args.get(i);
			Type argType = arg.type();

			if (!argType.isNull())
			{
				if (argType.isSubtype(SJConstants.SJSOCKET_TYPE))
				{
					if (!(arg instanceof Local))
					{
						throw new SemanticException("Only local session variables can be passed as method arguments: " + call);
					}

					SJLocalSocket sv = sjnf.SJLocalSocket(arg.position(), ((Local) arg).id());

					for (Object sjarg : sjargs)
					{
						if (sjarg instanceof SJLocalSocket && ((SJLocalSocket) sjarg).name().equals(sv.name())) // Could check this in type/alias checker.
						{
							throw new SemanticException("Illegal alias of SJSocket variable: " + call);
						}
					}

					sjargs.add(sv);

					isSJCall = true;
				}
				else if (argType.isSubtype(SJConstants.SJCHANNEL_TYPE) || argType.isSubtype(SJConstants.SJSERVER_TYPE))
				{
					throw new SemanticException("Argument passing of session server sockets not allowed: " + call);
				}
				else
				{
					sjargs.add(arg);
				}
			}
		}

		Type returnType = call.methodInstance().returnType();

		if (returnType.isSubtype(SJConstants.SJSOCKET_TYPE))
		{
			isSJCall = true;
		}
		else if (returnType.isSubtype(SJConstants.SJCHANNEL_TYPE) || returnType.isSubtype(SJConstants.SJSERVER_TYPE))
		{
			throw new SemanticException("Only session socket variables can be return by a session-method: " + call);
		}

		if (isSJCall)
		{
			boolean targetImplicit = call.isTargetImplicit();

			SJCall sjcall = sjnf.SJCall(call.position(), call.target(), call.id(), sjargs);

			sjcall = (SJCall) sjtb.buildTypes(context(), sjcall);
			sjcall = (SJCall) sjcall.targetImplicit(targetImplicit); // Hacky? (This value is originally set by the base type checker, not the type builder).

			if (returnType.isSubtype(SJConstants.SJSOCKET_TYPE))
			{
				if (!(parent instanceof Return)) // Will be parsed as a SJReturn.
				{
					String returned = SJCompilerUtils.parseSJReturnName(parent);

					sjcall = SJCompilerUtils.setSJCallReturnName(sjcall, returned);
				}
			}

			SJCompilerUtils.debugPrintln("Parsed: " + sjcall);

			call = sjcall;
		}

		return call;
	}

	private SJSpawn parseSJSpawn(SJSpawn sjs) throws SemanticException
	{
		New target = (New) sjs.target();

		if (!target.type().isSubtype(SJConstants.SJTHREAD_TYPE)) // Check in session type checker?
		{
			throw new SemanticException("SJThread required for session-spawn: " + sjs);
		}

		List<Receiver> socketArgs = new ArrayList<Receiver>();

		for (Object socket : SJCompilerUtils.getSJSpawnArgs(sjs))
		{
			SJSocketVariable sv = null;

			if (!(socket instanceof Local))
			{
				throw new SemanticException("Session-spawn operates on local session variables: " + sjs);
			}

			Local local = (Local) socket;

			if (!local.type().isSubtype(SJConstants.SJSOCKET_TYPE))
			{
				throw new SemanticException("Only session-sockets can be used in spawn: " + sjs);
			}

			sv = (SJLocalSocket) sjtb.buildTypes(context(), sjnf.SJLocalSocket(local.position(), local.id()));

			socketArgs.add(sv);
		}

		sjs = SJCompilerUtils.setSJSpawnArgs(sjs, socketArgs);

		List<Expr> args = new ArrayList<Expr>();

		args.addAll((List<Expr>) target.arguments());

		for (Receiver s : socketArgs)
		{
			args.add((Expr) s);
		}

		sjs = sjs.target((New) target.arguments(args));

		SJCompilerUtils.debugPrintln("Parsed: " + sjs); // Not very pretty.

		return sjs;
	}
}
