/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import java.util.*;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.Position;
import polyglot.visit.NodeVisitor;

import sj.SJConstants;
import sj.ast.*;
import sj.ast.protocoldecls.*;
import sj.ast.typenodes.*;
import sj.ast.createops.*;
import sj.ast.sessvars.*;
import sj.ast.sessops.basicops.*;
import sj.ast.sessops.structuralops.*;
import sj.ast.servops.*;
import sj.ast.sesscasts.*;
import sj.ast.sessformals.*;
import sj.ast.proceduredecls.*;
import sj.types.*;
import sj.types.typeobjects.sesstypes.*;
import sj.types.typeobjects.typeableinstances.*;
//import sj.visit.contexts.*;
import sj.util.*;

import sj.util.SJCompilerUtils;

public class SJSessionTypeBuilder extends SJContextVisitor
{
	private SJTypeSystem sjts = typeSystem();
	private SJNodeFactory sjnf = nodeFactory();

	//private SJTypeBuilder sjtb = new SJTypeBuilder(job(), sjts, sjnf);

	private SJTypeDisambiguator sjstd = new SJTypeDisambiguator(job(), sjts, sjnf, sjContext());

	public SJSessionTypeBuilder(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);
	}

	protected SJContextVisitor sjEnterCall(Node parent, Node n) throws SemanticException
	{
		return this;
	}

	protected void enterSJSessionTryScope(SJSessionTry sjst) throws SemanticException
	{
		sjContext().pushTry();
	}

	protected void enterSJStructuralScope(SJStructuralOperation sjso) throws SemanticException
	{
		sjContext().pushCode();
	}

	protected Node sjLeaveCall(Node parent, Node old, Node n, NodeVisitor v) throws SemanticException
	{
		if (n instanceof SJProtocolDecl) // Includes SJFormal. Separate out?
		{
			n = buildSJProtocolDecl((SJProtocolDecl) n);
		}
		else if (n instanceof SJCreateOperation)
		{
			n = buildSJCreateOperation((SJCreateOperation) n);
		}
		else if (n instanceof SJSessionTry)
		{
			n = buildSJSessionTry((SJSessionTry) n);
		}
		else if (n instanceof SJVariable)
		{
			n = buildSJVariable((SJVariable) n);
		}
		else if (n instanceof SJBasicOperation)
		{
			n = buildSJBasicOperation(parent, (SJBasicOperation) n);
		}
		else if (n instanceof SJCast)
		{
			n = buildSJCast((SJCast) n);
		}
		else if (n instanceof SJStructuralOperation)
		{
			n = buildSJStructualOperation((SJStructuralOperation) n);
		}
		else if (n instanceof SJServerOperation)
		{
			n = buildSJServerOperation((SJServerOperation) n);
		}
		/*else if (n instanceof SJCall)
		{
			// Can't look up the target method instance here, the session argument types are dynamic information: can only do it in SJTypeChecker. // Also, no instance object for recording sessions delegated/received types. // Similar for SJSpawn.
		}*/
		else if (n instanceof SJProcedureDecl)
		{
			/*if (n instanceof SJConstructorDecl)
			{

			}
			else */if (n instanceof SJMethodDecl)
			{
				n = buildSJMethodDecl((SJMethodDecl) n);
			}
		}
		else if (n instanceof SJFormal)
		{
			n = buildSJFormal((SJFormal) n);
		}
		else if (n instanceof SJReturn)
		{
			n = buildSJReturn((SJReturn) n);
		}

		return n;
	}

	protected SJSessionTry leaveSJSessionTryScope(SJSessionTry sjst) throws SemanticException // We are also building the instance object for session-trys in sjLeaveCall, but that routine does not modify the context stack, which is what we do here.
	{
		sjContext().pop();

		return sjst;
	}

	protected SJStructuralOperation leaveSJStructuralScope(SJStructuralOperation sjso) throws SemanticException
	{
		sjContext().pop();

		return sjso;
	}

	private SJProtocolDecl buildSJProtocolDecl(SJProtocolDecl pd) throws SemanticException
	{
		if (pd instanceof SJFieldProtocolDecl && !((SJFieldProtocolDecl) pd).flags().isPrivate())
		{
			throw new SemanticException("Non-private protocol field declarations not done yet: " + pd);
		}

		SJTypeNode sjTypeNode = sjstd.visit(context(), SJCompilerUtils.getSJProtocolDeclTypeNode(pd));

		SJType sjType = sjTypeNode.type();

		pd = SJCompilerUtils.setSJProtocolDeclTypeNode(pd, sjTypeNode);

		String protocol = SJCompilerUtils.getSJNamedName(pd); // Name is automatically synchronised with the AST node name value by the extension object getter/setter.
		SJNamedInstance sjni = sjts.SJNamedInstance(pd.position(), sjType, protocol);

		pd = (SJProtocolDecl) SJCompilerUtils.setSJNamedInstance(pd, sjni);

		sjContext().addProtocol(sjni); // Instance objects should be cloned?

		SJCompilerUtils.debugPrintln("Built: `" + protocol + "' " + sjType);

		return pd;
	}

	private SJCreateOperation buildSJCreateOperation(SJCreateOperation sjco) throws SemanticException
	{
		Position pos = sjco.position();

		SJType sjType = null;

		if (sjco instanceof SJChannelCreate || sjco instanceof SJServerCreate)
		{
			Expr arg = (Expr) sjco.arguments().get(0); // Factor out constant.

			sjType = sjContext().protocolType(resolveNamedVariableName(arg));

			if (!sjType.wellFormed())
			{
				throw new SemanticException("Protocol type not well-formed: " + sjType);
			}
		}
		else if (sjco instanceof SJSocketCreate)
		{
			Expr arg = (Expr) sjco.arguments().get(0);

			try // Using a received channel received with unknown type (SJTypeChecker will infer).
			{
				sjType = sjContext().channelType(resolveNamedVariableName(arg));
			}
			catch (SemanticException se)
			{
				sjType = sjts.SJUnknownType();
			}
		}

		String name = SJCompilerUtils.getSJNamedName(sjco);
		SJNamedInstance sjni = sjts.SJNamedInstance(pos, sjType, name);

		sjco = (SJCreateOperation) SJCompilerUtils.setSJNamedInstance(sjco, sjni);

		if (sjco instanceof SJChannelCreate)
		{
			sjContext().addChannel(sjni);

			SJCompilerUtils.debugPrintln("SJChannelCreate: `" + name + "' " + sjType);
		}
		else if (sjco instanceof SJSocketCreate)
		{
			sjContext().addSocket(sjni);

			SJCompilerUtils.debugPrintln("SJSocketCreate: `" + name + "' " + sjType);
		}
		else if (sjco instanceof SJServerCreate)
		{
			sjContext().addServer(sjni);

			SJCompilerUtils.debugPrintln("SJServerCreate: `" + name + "' " + sjType);
		}

		return sjco;
	}

	private String resolveNamedVariableName(Expr arg) throws SemanticException // Move to SJCompilerUtils/SJTypeSystem?
	{
		String name = null;

		if (arg instanceof Local)
		{
			name = ((Local) arg).name();
		}
		else if (arg instanceof Field)
		{
			name = ((Field) arg).name();
		}
		else
		{
			throw new SemanticException("Argument should be a named variable: " + arg); // Maybe relax this: for e.g. nested create calls.
		}

		return name;
	}

	private SJSessionTry buildSJSessionTry(SJSessionTry sjst)
	{
		SJMap sockets = new SJMap();

		for (Receiver sock : sjst.sockets())
		{
			String name = ((SJSocketVariable) sock).name();

			SJType sjt;

			try
			{
				sjt = sjContext().socketType(name);
			}
			catch (SemanticException se) // Socket has not been explicitly created.
			{
				sjt = sjts.SJUnknownType(); // Session will be accepted or received later.
			}

			sockets.put(name, sjt);
		}

		return SJCompilerUtils.setSJSessionTryInstance(sjst, sjts.SJSessionTryInstance(sjst.position(), sockets));
	}

	private SJVariable buildSJVariable(SJVariable sjv)
	{
		SJNamedInstance sjni = sjts.SJNamedInstance(sjv.position(), null, sjv.name()); // Session type component of instance object unused. SJTypeChecker could record these types for each operation, but not useful.

		return (SJVariable) SJCompilerUtils.setSJNamedInstance(sjv, sjni);
	}

	private SJBasicOperation buildSJBasicOperation(Node parent, SJBasicOperation sjbo) throws SemanticException
	{
		SJType sjType = null;

		if (sjbo instanceof SJRequest)
		{
			sjType = sjts.SJBeginType();
		}
		else if (sjbo instanceof SJSend)
		{
			Type messageType = (Type) ((Expr) sjbo.arguments().get(0)).type(); // Factor out constant?

			sjType = sjts.SJSendType(messageType);
		}
		else if (sjbo instanceof SJSendBoolean)
		{
			sjType = sjts.SJSendType(sjts.Boolean());
		}
		else if (sjbo instanceof SJSendInt)
		{
			sjType = sjts.SJSendType(sjts.Int());
		}
		else if (sjbo instanceof SJSendByte)
		{
			sjType = sjts.SJSendType(sjts.Byte());
		}
		else if (sjbo instanceof SJSendSession)
		{
			sjType = sjts.SJSendType(sjts.SJUnknownType());
		}
		else if (sjbo instanceof SJSendChannel)
		{
			Expr arg = (NamedVariable) sjbo.arguments().get(0); // Factor out constant? // Session operation parser checks for NamedVariable.

			sjType = sjts.SJSendType(sjContext().channelType(((NamedVariable) arg).name()));
		}
		else if (sjbo instanceof SJReceive)
		{
			if (parent instanceof Cast)
			{
				sjType = sjts.SJReceiveType(((Cast) parent).type());
			}
			else
			{
				sjType = sjts.SJReceiveType(sjts.SJUnknownType());
			}
		}
		else if (sjbo instanceof SJReceiveBoolean)
		{
			sjType = sjts.SJReceiveType(sjts.Boolean());
		}
		else if (sjbo instanceof SJReceiveInt)
		{
			sjType = sjts.SJReceiveType(sjts.Int());
		}
		else if (sjbo instanceof SJReceiveByte)
		{
			sjType = sjts.SJReceiveType(sjts.Byte());
		}
		else if (sjbo instanceof SJReceiveSession)
		{
			SJUnknownType sjut = sjts.SJUnknownType();

			if (!(parent instanceof SJCast)) // Partially duplicated in SJCast case.
			{
				SJReceiveSession sjrs = (SJReceiveSession) sjbo;
				String sess = SJCompilerUtils.getSJReturnName(sjrs);
				SJNamedInstance sjni = sjts.SJNamedInstance(sjrs.position(), sjut, sess);

				sjbo = (SJBasicOperation) SJCompilerUtils.setSJReturnOpInstance(sjrs, sjni);

				sjContext().addSocket(sjni); // Any point to this?
			}

			sjType = sjts.SJReceiveType(sjut); // Will be overwritten if casted.
		}
		else if (sjbo instanceof SJReceiveChannel)
		{
			SJUnknownType sjut = sjts.SJUnknownType();

			if (!(parent instanceof SJCast)) // Partially duplicated in SJCast case.
			{
				SJReceiveChannel sjrc = (SJReceiveChannel) sjbo;
				String chan = SJCompilerUtils.getSJReturnName(sjrc);
				SJNamedInstance sjni = sjts.SJNamedInstance(sjrc.position(), sjut, chan);

				sjbo = (SJBasicOperation) SJCompilerUtils.setSJReturnOpInstance(sjrc, sjni);

				sjContext().addChannel(sjni);
			}

			sjType = sjts.SJReceiveType(sjts.SJUnknownType()); // Will be overwritten if casted.
		}
		else if (sjbo instanceof SJRecurse)
		{
			sjType = sjts.SJRecurseType(((SJRecurse) sjbo).label());
		}

		SJSessionOpInstance sjsoi = sjts.SJSessionOpInstance(sjbo.position(), sjType, sjbo.target().name());

		sjbo = (SJBasicOperation) SJCompilerUtils.setSJSessionOpInstance(sjbo, sjsoi);

		SJCompilerUtils.debugPrintln("Built: `" + sjbo + "' " + sjType);

		return sjbo;
	}

	private SJBasicOperation buildSJCast(SJCast sjc) throws SemanticException
	{
		SJTypeCast sjtc = (SJTypeCast) sjc;
		SJBasicOperation sjbo = (SJBasicOperation) sjtc.expr(); // Channel/session receive.

		SJTypeNode sjTypeNode = sjstd.visit(context(), SJCompilerUtils.getSJTypeCastTypeNode(sjtc));

		SJType sjType = sjTypeNode.type();

		SJNamedInstance sjni  = null;

		if (sjbo instanceof SJReceiveChannel)
		{
			if (!sjType.wellFormed())
			{
				throw new SemanticException("Received channel type not well-formed: " + sjType);
			}

			SJReceiveChannel sjrc = (SJReceiveChannel) sjbo;
			String chan = SJCompilerUtils.getSJReturnName(sjrc);

			sjni = sjts.SJNamedInstance(sjrc.position(), sjType, chan);
			sjbo = (SJBasicOperation) SJCompilerUtils.setSJReturnOpInstance(sjrc, sjni);

			sjContext().addChannel(sjni);
		}
		else //if (sjbo instanceof SJReceiveSession)
		{
			if (!sjType.treeWellFormed()) // Includes check for bad 'begin' elements.
			{
				throw new SemanticException("Received session type not well-formed: " + sjType);
			}

			SJReceiveSession sjrs = (SJReceiveSession) sjbo;
			String sess = SJCompilerUtils.getSJReturnName(sjrs);

			sjni = sjts.SJNamedInstance(sjrs.position(), sjType, sess);
			sjbo = (SJBasicOperation) SJCompilerUtils.setSJReturnOpInstance(sjrs, sjni);

			sjContext().addSocket(sjni);
		}

		sjType = sjts.SJReceiveType(sjType);

		SJSessionOpInstance sjsoi = sjts.SJSessionOpInstance(sjbo.position(), sjType, sjbo.target().name());

		sjbo = (SJBasicOperation) SJCompilerUtils.setSJSessionOpInstance(sjbo, sjsoi);

		SJCompilerUtils.debugPrintln("Built: `" + sjbo + "' " + sjType);

		return sjbo; // Drop the cast, no longer needed.
	}

	private SJStructuralOperation buildSJStructualOperation(SJStructuralOperation sjso)
	{
		String session = ((SJSocketVariable) SJCompilerUtils.getSJStructuralOpTarget(sjso)).name();

		SJType sjType = null;

		if (sjso instanceof SJOutbranch) // Could visit structural operation contexts recursively but not really needed.
		{
			sjType = sjts.SJOutbranchType().branchCase(((SJOutbranch) sjso).label(), sjts.SJUnknownType());
		}
		else if (sjso instanceof SJInbranch)
		{
			sjType = sjts.SJInbranchType();
		}
		else if (sjso instanceof SJOutwhile)
		{
			sjType = sjts.SJOutwhileType().body(sjts.SJUnknownType());
		}
		else if (sjso instanceof SJInwhile)
		{
			sjType = sjts.SJInwhileType().body(sjts.SJUnknownType());
		}
		else if (sjso instanceof SJRecursion)
		{
			sjType = sjts.SJRecursionType(((SJRecursion) sjso).label()).body(sjts.SJUnknownType());
		}

		SJSessionOpInstance sjsoi = sjts.SJSessionOpInstance(sjso.position(), sjType, session);

		sjso = (SJStructuralOperation) SJCompilerUtils.setSJSessionOpInstance(sjso, sjsoi);

		//SJCompilerUtils.debugPrintln("Built: `" + sjso + "' " + sjType); // Structural operations don't print properly: need to override the appropriate methods in the AST nodes.
		SJCompilerUtils.debugPrintln("Built: `" + session + "' " + sjType);

		return sjso;
	}

	private SJServerOperation buildSJServerOperation(SJServerOperation sjso) throws SemanticException
	{
		String server = SJCompilerUtils.getSJNamedName(sjso);
		Position pos = sjso.position();

		SJType sjType = null;

		if (sjso instanceof SJAccept)
		{
			SJAccept sja = (SJAccept) sjso;
			String returned = SJCompilerUtils.getSJReturnName(sja);

			sjType = sjContext().serverType(server);

			sjso = (SJAccept) SJCompilerUtils.setSJReturnOpInstance(sja, sjts.SJNamedInstance(pos, sjType.child(), returned)); // Currently unused.
		}

		sjso = (SJServerOperation) SJCompilerUtils.setSJNamedInstance(sjso, sjts.SJNamedInstance(pos, sjType, server));

		SJCompilerUtils.debugPrintln("Built: `" + sjso + "' " + sjType);

		return sjso;
	}

	private SJMethodDecl buildSJMethodDecl(SJMethodDecl sjmd) throws SemanticException
	{
		if (!sjmd.flags().isPrivate()) // Should also make final (maybe when private is relaxed)? // Couldn't check this is SJProcedureDeclParser because no type information available.
		{
			if (!(sjmd.name().equals(SJConstants.SJTHREAD_RUN) && context().currentClass().isSubtype(SJConstants.SJTHREAD_TYPE)))
			{
				throw new SemanticException("Non-private session-methods not done yet: " + sjmd);
			}
		}

		String name = sjmd.name();

		SJMap sjFormalsMap = new SJMap();

		List<SJFormal> sjFormals = SJCompilerUtils.getSJMethodDeclFormals(sjmd);

		//for (SJFormal sjFormal : SJCompilerUtils.getSJMethodDeclFormals(sjmd))
		for (int i = 0; i < sjFormals.size(); i++) // To guarantee ordering?
		{
			SJFormal sjFormal = sjFormals.get(i);

			SJType sjType = null;

			if (sjFormal instanceof SJTypeFormal) // Factor out?
			{
				sjType = SJCompilerUtils.getSJTypeFormalTypeNode((SJTypeFormal) sjFormal).type();
			}

			if (sjmd instanceof SJRecursiveMethodDecl)
			{
				/*if (!(sjType instanceof SJRecursionType)) // Safe to relax this?
				{
					throw new SemanticException("Recursive session methods can only accept recursion type arguments: " + sjmd);
				}
				else*/ if (sjType instanceof SJRecursionType && !((SJRecursionType) sjType).label().labelValue().equals(sjmd.name()))
				{
					throw new SemanticException("Recursion argument type does not match method name: " + sjmd);
				}
			}

			sjFormalsMap.put(sjFormal.name(), sjType); // Ordering preserved by SJMap.
		}

		SJTypeNode returnTypeNode = SJCompilerUtils.getSJMethodDeclReturnType(sjmd);
		SJType sjReturnType = null;

		if (returnTypeNode != null)
		{
			returnTypeNode = sjstd.visit(context(), returnTypeNode);

			sjReturnType = (returnTypeNode == null) ? null : returnTypeNode.type(); // Needs to be disambiguated?

			sjmd = SJCompilerUtils.setSJMethodDeclReturnType(sjmd, returnTypeNode);
		}

		SJMethodInstance sjmi = null;

		if (sjmd instanceof SJRecursiveMethodDecl)
		{
			sjmi = sjts.SJRecursiveMethodInstance(name, sjFormalsMap, sjReturnType);
		}
		else
		{
			sjmi = sjts.SJMethodInstance(name, sjFormalsMap, sjReturnType);
		}

		sjmd = SJCompilerUtils.setSJMethodInstance(sjmd, sjmi); // Could add the method to the current class context (need a special context operation to do this, to access the outer class context).

		SJCompilerUtils.debugPrintln("Built: " + name + ", " + sjFormalsMap + ", " + sjReturnType);

		return sjmd;
	}

	private SJFormal buildSJFormal(SJFormal sjFormal)
	{
		if (sjFormal instanceof SJTypeFormal)
		{
			SJTypeFormal sjtf = (SJTypeFormal) sjFormal;

			SJTypeNode sjTypeNode = sjstd.visit(context(), SJCompilerUtils.getSJTypeFormalTypeNode(sjtf));

			SJType sjType = sjTypeNode.type();

			sjtf = SJCompilerUtils.setSJTypeFormalTypeNode(sjtf, sjTypeNode);

			String name = SJCompilerUtils.getSJNamedName(sjtf);
			SJNamedInstance sjni = sjts.SJNamedInstance(sjtf.position(), sjType, name);

			sjtf = (SJTypeFormal) SJCompilerUtils.setSJNamedInstance(sjtf, sjni);

			sjContext().addSocket(sjni);

			SJCompilerUtils.debugPrintln("Built: `" + name + "' " + sjType);
		}

		return sjFormal;
	}

	private SJReturn buildSJReturn(SJReturn sjr) throws SemanticException
	{
		SJType sjType = sjts.SJSendType(sjts.SJUnknownType()); // Can't yet look up the return type of the current session-method context because that information has only just been built this pass (and a dummy context was pushed instead). This will be done in SJTypeChecker.
		String session = SJCompilerUtils.getSJNamedName(sjr); // Null if returning a SJCall.

		sjr = (SJReturn) SJCompilerUtils.setSJNamedInstance(sjr, sjts.SJNamedInstance(sjr.position(), sjType, session));

		SJCompilerUtils.debugPrintln("Built: `" + sjr + "' " + sjType);

		return sjr;
	}
}
