/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import java.util.*;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.ErrorHandlingVisitor;
import polyglot.visit.NodeVisitor;

import sj.SJConstants;
import sj.ast.*;
import sj.ast.sessvars.SJSocketVariable;
import sj.ast.sessops.basicops.SJRecurse;
import sj.ast.sessops.structuralops.*;
import sj.ast.socketops.SJSocketOperation;
import sj.types.SJTypeSystem;
import sj.util.*;

/**
 * Also translates basic recurse operation.
 */
public class SJStructuralOperationTranslator extends ErrorHandlingVisitor
{
	private SJTypeSystem sjts = (SJTypeSystem) typeSystem();
	private SJNodeFactory sjnf = (SJNodeFactory) nodeFactory();

	private SJCounter sjc = new SJCounter();

	public SJStructuralOperationTranslator(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);
	}

	protected Node leaveCall(Node old, Node n, NodeVisitor v) throws SemanticException
	{
		if (n instanceof SJStructuralOperation)
		{
			SJStructuralOperation sjso = (SJStructuralOperation) n;

			Position pos = sjso.position();
			SJSocketVariable sv = (SJSocketVariable) SJCompilerUtils.getSJStructuralOpTarget(sjso);

			if (sjso instanceof SJOutbranch)
			{
				SJSocketOperation sockop = sjnf.SJOutlabel(pos, sv, ((SJOutbranch) sjso).label());
				//sockop = (SJSocketOperation) sjtb.buildTypes(context(), sockop);

				n = ((SJOutbranch) sjso).prepend(sjnf.Eval(sjso.position(), sockop));
			}
			else if (sjso instanceof SJInbranch)
			{
				SJSocketOperation sockop = sjnf.SJInlabel(pos, sv, ((SJInbranch) sjso).arguments());

				n = translateSJInbranch((SJInbranch) sjso, sockop);
			}
			else if (sjso instanceof SJOutwhile)
			{
				Expr cond = (Expr) ((Call) ((SJOutwhile) sjso).cond()).arguments().get(0); // Find a better way to carry arguments over from original parse pass. // Should check correct argument arity?

				SJSocketOperation sockop = sjnf.SJOutsync(pos, sv, cond);

				n = ((SJOutwhile) sjso).cond(sockop);
			}
			else if (sjso instanceof SJInwhile)
			{
				SJSocketOperation sockop = sjnf.SJInsync(pos, sv, (List<Expr>) ((Call) ((SJInwhile) sjso).cond()).arguments());

				n = ((SJInwhile) sjso).cond(sockop);
			}
			else if (sjso instanceof SJRecursion)
			{
				SJSocketOperation sockop = sjnf.SJRecursionFalse(pos, sv, ((SJRecursion) sjso).label());

				n = translateSJRecursion((SJRecursion) sjso, sockop);
			}
		}
		else if (n instanceof SJRecurse)
		{
			SJRecurse sjr = (SJRecurse) n;
			String sessionName = SJCompilerUtils.getSessionName(sjr);
			SJLabel lab = sjr.label();
			Position pos = sjr.position();

			SJSocketOperation sockop = sjnf.SJRecursionTrue(pos, sjr.target(), lab);

			Id id = sjnf.Id(pos, SJConstants.SJ_RECURSION_PREFIX + sessionName + "_" + lab.labelValue());

			n = sjnf.LocalAssign(pos, sjnf.Local(pos, id), Assign.ASSIGN, sockop);
		}

		return n;
	}

	private Block translateSJInbranch(SJInbranch sjib, SJSocketOperation sockop) // sockop is a SJInlabel.
	{
		Position pos = sjib.position();

		String branchValue = SJConstants.SJ_INBRANCH_LABEL_FIELD_PREFIX + sjc.nextValue();

		LocalDecl ld = sjnf.LocalDecl(pos, Flags.NONE, sjnf.CanonicalTypeNode(pos, SJConstants.SJLABEL_TYPE), sjnf.Id(pos, branchValue), sockop);

		SJInbranchCase[] cases = (SJInbranchCase[]) sjib.branchCases().toArray(new SJInbranchCase[0]);

		Stmt ifCases = null;

		int redo = cases.length - 1;

		//for (int i = 0; i < cases.length; i++)
		for (int i = redo; i >= 0 ; i--)
		{
			if (i == redo) // First iteration.
			{
				ifCases = sjnf.Block(pos, cases[redo]);
			}
			else
			{
				SJLabel lab = cases[i].label();

				ifCases = sjnf.If(pos, sjnf.Call(pos, sjnf.Local(pos, sjnf.Id(pos, branchValue)), sjnf.Id(pos, SJConstants.SJLABEL_EQUALITY_TEST), sjnf.StringLit(pos, lab.labelValue())), cases[i]).alternative(ifCases);
			}
		}

		//List l = new TypedList(new LinkedList(), Stmt.class, false);
		List<Stmt> l = new LinkedList<Stmt>();

		l.add(ld);
		l.add(ifCases);

		return sjnf.Block(pos, l);
	}

	private For translateSJRecursion(SJRecursion sjr, SJSocketOperation sockop)
	{
		String session = SJCompilerUtils.getSessionName(sjr);
		//SJLabel lab = sjr.label();
		Position pos = sjr.position();

		Id id = sjnf.Id(pos, SJConstants.SJ_RECURSION_PREFIX + session + "_" + sjr.label().labelValue());

		List<ForInit> inits = new LinkedList<ForInit>();
		List<ForUpdate> iters = new LinkedList<ForUpdate>();

		inits.add(sjnf.LocalDecl(pos, Flags.NONE, sjnf.CanonicalTypeNode(pos, sjts.Boolean()), id, sjnf.BooleanLit(pos, true)));

		Block translated = (Block) sjr.prepend(sjnf.Eval(pos, sjnf.LocalAssign(pos, sjnf.Local(pos, id), Assign.ASSIGN, sockop)));

		return sjnf.For(pos, inits, sjnf.Local(pos, id), iters, translated);
	}
}
