/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import java.util.*;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
//import polyglot.util.TypedList;
import polyglot.visit.ContextVisitor;
import polyglot.visit.NodeVisitor;

import sj.ast.typenodes.*;
import sj.types.SJTypeSystem;
import sj.types.typeobjects.sesstypes.*;
import sj.util.*;
import sj.visit.contexts.*;

/**
 * SJTypeDisambiguator builds the types for protocol declarations. Not actually run as a compiler pass, but run on demand by SJTypeBuilder.
 */
public class SJTypeDisambiguator extends ContextVisitor
{
	private SJTypeSystem sjts = (SJTypeSystem) typeSystem();

	private SJTypeBuilder sjtb = new SJTypeBuilder(job(), sjts, nodeFactory());

	private SJContext sjc;

	public SJTypeDisambiguator(Job job, TypeSystem ts, NodeFactory nf, SJContext sjc)
	{
		super(job, ts, nf); // Hacky for job to be fixed at construction but not the sjvc?

		this.sjc = sjc;
	}

	public SJTypeNode visit(Context context, SJTypeNode sjTypeNode)
	{
		this.context = (Context) context.copy();

		sjTypeNode = (SJTypeNode) sjTypeNode.visit(this);

		SJCompilerUtils.debugPrintln("Disambiguated: " + sjTypeNode + " " + sjTypeNode.type());

		return sjTypeNode;
	}

	protected NodeVisitor enterCall(Node parent, Node n) throws SemanticException
	{
		return this;
	}

	protected Node leaveCall(Node old, Node n, NodeVisitor v) throws SemanticException
	{
		if (n instanceof SJTypeNode) // Should always be true - called on demand. // Following routines could be moved into the type nodes.
		{
			SJTypeNode sjtn = (SJTypeNode) n;

			if (sjtn.child() != null)
			{
				sjtn = sjtn.child((SJTypeNode) sjtn.child().visit(this));
			}

			SJType sjType = null;

			if (sjtn instanceof SJBeginNode)
			{
				sjType = sjts.SJBeginType();
			}
			else if (sjtn instanceof SJSendNode)
			{
				SJSendNode sn = (SJSendNode) sjtn;

				TypeNode messageTypeNode = sn.messageTypeNode();

				if (messageTypeNode instanceof SJTypeNode)
				{
					if (messageTypeNode instanceof SJProtocolNode)
					{
						messageTypeNode = messageTypeNode.type(lookupProtocol((SJProtocolNode) messageTypeNode)); // Cloned by lookup operation.
					}
					else
					{
						messageTypeNode = (TypeNode) messageTypeNode.visit(this);
					}
				}
				else
				{
					messageTypeNode = (TypeNode) sjtb.buildTypes(context(), messageTypeNode);
				}

				n = sn.messageTypeNode(messageTypeNode); // Would be better to have the visitors involved (TypeBuilder, etc.) visit these nodes as part of their original passes.

				sjType = sjts.SJSendType(messageTypeNode.type());
			}
			else if (sjtn instanceof SJReceiveNode)
			{
				SJReceiveNode rn = (SJReceiveNode) sjtn;

				TypeNode messageTypeNode = rn.messageTypeNode();

				if (messageTypeNode instanceof SJTypeNode)
				{
					if (messageTypeNode instanceof SJProtocolNode)
					{
						messageTypeNode = messageTypeNode.type(lookupProtocol((SJProtocolNode) messageTypeNode)); // Cloned by lookup operation.
					}
					else
					{
						messageTypeNode = (TypeNode) messageTypeNode.visit(this);
					}
				}
				else
				{
					messageTypeNode = (TypeNode) sjtb.buildTypes(context(), messageTypeNode);
				}

				n = rn.messageTypeNode(messageTypeNode);

				sjType = sjts.SJReceiveType(messageTypeNode.type());
			}
			else if (sjtn instanceof SJOutbranchNode)
			{
				SJOutbranchNode obn = (SJOutbranchNode) sjtn;

				List<SJBranchCaseNode> branchCases = obn.branchCases();
				//List<SJBranchCaseNode> visitedCases = new TypedList(new LinkedList(), SJBranchCaseNode.class, false); // Unnecessary to use TypedList, since using generics (but uniform with Polyglot?)
				List<SJBranchCaseNode> visitedCases = new LinkedList<SJBranchCaseNode>();

				SJOutbranchType obt = sjts.SJOutbranchType();

				for (SJBranchCaseNode bcn : branchCases)
				{
					SJTypeNode sjTypeNode = bcn.sjTypeNode();

					if (sjTypeNode != null) // Inbranch cases may be null.
					{
						sjTypeNode = (SJTypeNode) sjTypeNode.visit(this);
					}

					visitedCases.add(bcn.sjTypeNode(sjTypeNode));

					obt.branchCase(bcn.label(), (sjTypeNode == null ? null : sjTypeNode.type()));
				}

				n = obn.branchCases(visitedCases);

				sjType = obt;
			}
			else if (sjtn instanceof SJInbranchNode)
			{
				SJInbranchNode ibn = (SJInbranchNode) sjtn;

				List<SJBranchCaseNode> branchCases = ibn.branchCases();
				//List<SJBranchCaseNode> visitedCases = new TypedList(new LinkedList(), SJBranchCaseNode.class, false);
				List<SJBranchCaseNode> visitedCases = new LinkedList<SJBranchCaseNode>();

				SJInbranchType ibt = sjts.SJInbranchType();

				for (SJBranchCaseNode bcn : branchCases)
				{
					SJTypeNode sjTypeNode = (SJTypeNode) bcn.sjTypeNode();

					if (sjTypeNode != null)
					{
						sjTypeNode = (SJTypeNode) sjTypeNode.visit(this);
					}

					visitedCases.add(bcn.sjTypeNode(sjTypeNode));

					ibt.branchCase(bcn.label(), (sjTypeNode == null ? null : sjTypeNode.type()));
				}

				n = ibn.branchCases(visitedCases);

				sjType = ibt;
			}
			else if (sjtn instanceof SJOutwhileNode)
			{
				SJOutwhileNode own = (SJOutwhileNode) sjtn;
				SJTypeNode body = (SJTypeNode) own.body();

				sjType = sjts.SJOutwhileType();

				if (body != null)
				{
					body = (SJTypeNode) body.visit(this);

					n = own.body(body);

					sjType = ((SJOutwhileType) sjType).body(body.type());
				}
			}
			else if (sjtn instanceof SJInwhileNode)
			{
				SJInwhileNode iwn = (SJInwhileNode) sjtn;
				SJTypeNode body = iwn.body();

				sjType = sjts.SJInwhileType();

				if (body != null)
				{
					body = (SJTypeNode) body.visit(this);

					n = iwn.body(body);

					sjType = ((SJInwhileType) sjType).body(body.type());
				}
			}
			else if (sjtn instanceof SJRecursionNode)
			{
				SJRecursionNode sjrn = (SJRecursionNode) sjtn;
				SJTypeNode body = (SJTypeNode) sjrn.body();

				sjType = sjts.SJRecursionType(sjrn.label());

				if (body != null)
				{
					body = (SJTypeNode) body.visit(this);

					n = sjrn.body(body);

					sjType = ((SJRecursionType) sjType).body(body.type());
				}
			}
			else if (sjtn instanceof SJRecurseNode)
			{
				sjType = sjts.SJRecurseType(((SJRecurseNode) sjtn).label());
			}
			else if (sjtn instanceof SJProtocolNode)
			{
				SJProtocolNode sjpn = (SJProtocolNode) sjtn;

				sjType = lookupProtocol(sjpn);
			}
			else
			{
				throw new RuntimeException("Shouldn't get here: " + sjtn);
			}

			n = sjtn.type(sjType);
		}

		return n;
	}

	private SJType lookupProtocol(SJProtocolNode pn) throws SemanticException
	{
		String name = pn.protocol();

		/*if (...)
		{
			// Need to prevent (mutual and self) recursion in protocol declarations. // Mutual currently prevented by single-pass linear context building (uniform with e.g. regular static final declarations).
		}*/

		SJType sjType = sjc.findProtocol(name).sjType();

		if (sjType == null)
		{
			throw new SemanticException("Unknown protocol: " + name); // Better if invalid protocol identifiers caught by base type checker. But should also allow field indirection (standard Java static finals do not however)?
		}

		return sjType.treeClone(); // In case of multiple references from the same protocol (won't be null).
	}
}
