/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.runtime.net;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.Map.Entry;

import polyglot.types.Type;

import sj.types.typeobjects.sesstypes.*;
import sj.util.*;

import sj.runtime.*;
import sj.runtime.util.*;

//import org.jboss.serial.io.*;

public final class SJRSocket extends SJSocket
{
	//BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
	//try { br.readLine(); } catch (IOException ioe) { }

	protected static final boolean RUNTIME_STATE = true; // Should only be disabled if no delegation.
	protected static final boolean CHECK_DUALITY = true;

	private static final int INIT_SER_BUF_SIZE = 32; // The Java default. // Doesn't seem to make much difference making this bigger? (If anything, it's slower?)

	private static final byte SJ_FIN = 0; // Distinguish success and failure FIN?

	//private static final byte SJ_OBJECT = 1;
	private static final byte SJ_BOOLEAN = 2;
	private static final byte SJ_INT = 3;
	private static final byte SJ_BYTE = 4;
	private static final byte SJ_SERIALIZED = 9;

	private final SJType protocolType;
	protected final SJSocketState state;

	private ObjectOutputStream oos;
	private ObjectInputStream ois;

	//private static final SJFIN FIN = new SJFIN();

	// For delegation.
	private static final byte SJ_START_DEL = 10;
	private static final byte SJ_DEL_SIGNAL = 11;
	private static final byte SJ_DEL_ACK = 12;

	//private static final Object DS_ACK = new SJDSAck();
	protected static final SJRSerializedMessage CACHE_MARKER = null; // Hacky weakened access.

	private final Stack<SJRSerializedMessage> sentMessageCache = new Stack<SJRSerializedMessage>(); // Why stack?

	// For simultaneous delegation.
	private static final byte SINGLE_DEL = 20;
	private static final byte SIM_DEL = 21;
	//private static final Object END_OF_RESEND = null; // Define a dedicated object type for this purpose.
	private static final byte SJ_END_OF_RESEND = 13;

	private final List<Object> lostMessages = new LinkedList<Object>(); // Unlike SJFSocket, this is only for double delegation.

	public final static SJRSocket create(SJServerAddress channel)
	{
		return new SJRSocket(channel);
	}

	private SJRSocket(SJServerAddress channel) // Almost the same as in SJFSocket.
	{
		setChannel(channel);

		this.protocolType = decodeSJType(channel.getProtocol());

		if (!RUNTIME_STATE)
		{
			this.state = null;
		}
		else
		{
			this.state = new SJSocketState(this, protocolType); // Runtime state tracking is mandatory for resending-based session sockets.
		}
	}

	protected SJRSocket(SJType sjType, SJProtocol encoded, Socket socket)
	{
		setChannel(SJServerAddress.create(encoded, (InetSocketAddress) socket.getRemoteSocketAddress()));
		setSocket(socket);

		this.protocolType = sjType.treeClone();

		if (!RUNTIME_STATE)
		{
			this.state = null;
		}
		else
		{
			this.state = new SJSocketState(this, protocolType);
		}
	}

	public final void request() throws UnknownHostException, SJIOException, SJIncompatibleSessionException // Almost the same as in SJFSocket.
	{
		checkClassLoadingParameters();

		connect(getChannel().targetAddress());
		init(); // Throws an SJException if any initiation problems.

		isActive(true);
		setOriginalRequestor(IS_ORIGINAL_REQUESTOR);

		if (RUNTIME_STATE)
		{
			state.request();

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}
	}

	private final void connect(InetSocketAddress addr) throws SJIOException
	{
		try
		{
			Socket socket = new Socket(addr.getAddress(), addr.getPort());

			socket.setTcpNoDelay(TCP_NO_DELAY);

			setSocket(socket);
			requestIOStreams();

			//SJRuntimeUtils.debugPrintln("Connected on port " + socket.getLocalPort() + " to: " + socket.getInetAddress().getHostName() + ":" + socket.getPort());
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	/*private BufferedOutputStream bos = null;
	private BufferedInputStream bis = null;*/

	private final void requestIOStreams() throws SJIOException
	{
		try
		{
			Socket s = getSocket();

			/*this.bos = new BufferedOutputStream(s.getOutputStream());
			this.bis = new BufferedInputStream(s.getInputStream());*/

			this.oos = new ObjectOutputStream(s.getOutputStream());

			if (remoteClassLoadingEnabled())
			{
				SJRemoteClassLoader sjrcl = new SJRemoteClassLoader(this.getClass().getClassLoader(), this);

				this.ois = new SJObjectInputStream(s.getInputStream(), sjrcl);

				setRemoteClassLoader(sjrcl);
			}
			else
			{
				this.ois = new ObjectInputStream(s.getInputStream());
			}
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	protected final void acceptIOStreams() throws SJIOException
	{
		try
		{
			Socket s = getSocket();

			if (remoteClassLoadingEnabled())
			{
				SJRemoteClassLoader sjrcl = new SJRemoteClassLoader(this.getClass().getClassLoader(), this);

				this.ois = new SJObjectInputStream(s.getInputStream(), sjrcl);

				setRemoteClassLoader(sjrcl);
			}
			else
			{
				this.ois = new ObjectInputStream(s.getInputStream());
			}

			this.oos = new ObjectOutputStream(s.getOutputStream());
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	public final void sendSession(SJSocket socket, String encoded) throws SJIOException
	{
		if (!RUNTIME_STATE)
		{
			throw new SJIOException("Runtime state tracking disabled, delegation not possible.");
		}

		if (!(socket instanceof SJRSocket))
		{
			throw new SJIOException("Cannot delegate a session socket with different implementation: " + socket.getClass());
		}

		SJRSocket rsocket = (SJRSocket) socket;

		try
		{
			rsocket.isActive(false);

			//if (!rsocket.state.currentContext().isTopLevelContext()) rsocket.state.advanceContext(); // Branch contexts are advanced by last operation in the context, in this case the delegation.

			SJType rsocketType = rsocket.currentState(); // Should correspond to `encoded'.

			int port = -1;

			SJStartDelegation sjsd = new SJStartDelegation(rsocket.isOriginalRequestor(), rsocket.getSocketParameters());

			while (true) // Similar to readIntWithDelegationCheck, but need to resend the start delegation each time we find a delegation signal (case 3).
			{
				writeByte(SJ_START_DEL);
				writeObject(sjsd);

				byte flag = readByte();

				if (flag == SJ_DEL_SIGNAL)
				{
					reconnectSingleDelegation((SJDelegationSignal) readObjectNoCNFE()); // FIN possible?
				}
				else //if (flag == SJ_INT)
				{
					port = readInt();

					break;
				}
			}

			//SJRuntimeUtils.debugPrintln("Forwarding target port: " + port);

			SJDelegationSignal ourds = new SJDelegationSignal(getTargetAddress().getHostName(), port, encodeSJType(rsocketType)); // Could just use `encoded'? // Current session type could instead be sent by session-receiver after connection is restablished (any advantage?).

			rsocket.writeByte(SJ_DEL_SIGNAL); // Not cached.
			rsocket.writeObject(ourds); // Follows writeSerialized protocol.

			SJType expected = rsocket.expectedType();

			Object ack = getDSReply(rsocket);

			if (ack instanceof SJStartDelegation)
			{
				// Nothing to do (except close the socket below).
			}
			else if (ack instanceof SJDelegationSignal) // Simultaneous delegation,
			{
				rsocket.reconnectSimultaneousDelegation((SJDelegationSignal) ack, ourds);
			}
			else
			{
				// ack should be null (for SJ_DEL_ACK or SJ_FIN).
			}

			state.sendSession(expected);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}
		finally
		{
			try
			{
				//SJRuntimeUtils.debugPrintln("Closing (" + rsocket.getLocalAddress().getPort() + "): " + rsocket.state.currentState());

				rsocket.state.close();

				rsocket.shutdown();
			}
			catch (SJIOException ioe) { } // Should this exception be supressed?
		}
	}

	private final Object getDSReply(SJRSocket rsocket) throws SJIOException // Basically executes the delegated session until we get an acknowledgement (implicit or explicit) to our delegation signal.
	{
		SJType expected = rsocket.expectedType();

		if (expected instanceof SJReceiveType) // `?' type means we expect a lost message (including start delegation) or a delegation signal (or a failure FIN). // Would be faster to just use flags and not actual type tracking?
		{
			Type messageType = ((SJReceiveType) expected).messageType();

			if (messageType.isClass() || messageType instanceof SJType) // Includes delegation cases 3.
			{
				byte flag = rsocket.readByte();

				if (flag == SJ_SERIALIZED)
				{
					try
					{
						rsocket.state.receive(rsocket.readSerialized());
					}
					catch (ClassNotFoundException cnfe)
					{
						//throw new SJIOException(cnfe);
					}

					return getDSReply(rsocket);
				}
				else if (flag == SJ_START_DEL)
				{
					return null;
				}
				else if (flag == SJ_DEL_SIGNAL)
				{
					return rsocket.readObjectNoCNFE(); // Could suppress ClassNotFoundExceptions, don't care if we can't deserialize lost messages.
				}
				else //if (flag == SJ_FIN)
				{
					throw new SJIOException("Session prematurely closed by peer.");
				}
			}
			else
			{
				if (messageType.isPrimitive())
				{
					byte flag = rsocket.readByte();

					if (flag == SJ_BOOLEAN)
					{
						rsocket.state.receiveBoolean(rsocket.readBoolean());

						return getDSReply(rsocket);
					}
					else if (flag == SJ_INT)
					{
						rsocket.state.receiveInt(rsocket.readInt());

						return getDSReply(rsocket);
					}
					else if (flag == SJ_BYTE)
					{
						rsocket.state.receiveByte(rsocket.readByte());

						return getDSReply(rsocket);
					}
					else if (flag == SJ_DEL_SIGNAL)
					{
						return rsocket.readObjectNoCNFE();
					}
					else //if (flag == SJ_FIN)
					{
						throw new SJIOException("Session prematurely closed by peer.");
					}
				}
				/*else if (messageType.isArray())
				{

				}*/
				else
				{
					throw new RuntimeException("Message type not handled: " + messageType);
				}
			}
		}
		else if (expected instanceof SJInbranchType)
		{
			byte flag = rsocket.readByte();

			if (flag == SJ_SERIALIZED) // Make a SJ_LABEL flag?
			{
				try
				{
					rsocket.state.inbranch(new SJLabel((String) rsocket.readSerialized()));
				}
				catch (ClassNotFoundException cnfe) { }

				return getDSReply(rsocket);
			}
			else if (flag == SJ_DEL_SIGNAL)
			{
				return rsocket.readObjectNoCNFE();
			}
			else //if (flag == SJ_FIN)
			{
				throw new SJIOException("Session prematurely closed by peer.");
			}
		}
		else if (expected instanceof SJInwhileType)
		{
			byte flag = rsocket.readByte(); // Dual to outsync.

			if (flag == SJ_BOOLEAN) // i.e. the outwhile boolean.
			{
				rsocket.state.inwhile(rsocket.readBoolean());

				return getDSReply(rsocket);
			}
			else if (flag == SJ_DEL_SIGNAL)
			{
				return rsocket.readObjectNoCNFE();
			}
			else //if (flag == SJ_FIN)
			{
				throw new SJIOException("Session prematurely closed by peer.");
			}
		}
		else
		{
			byte flag = rsocket.readByte();

			if (flag == SJ_DEL_SIGNAL)
			{
				return rsocket.readObjectNoCNFE();
			}
			else if (flag == SJ_DEL_ACK)
			{
				return null;
			}
			else //if (flag == SJ_FIN) // Don't know if passive party failed or actually finished.
			{
				//throw new SJIOException("Session prematurely closed by peer.");

				return null;
			}
		}
	}

	public final SJRSocket receiveSession(String encoded) throws UnsupportedOperationException, SJIOException
	{
		if (!RUNTIME_STATE)
		{
			throw new SJIOException("Runtime state tracking disabled, delegation not possible.");
		}

		SJProtocol protocol = new SJProtocol(encoded);

		int port = SJRuntimeUtils.findFreeTCPPort();

		ServerSocket ss = null;
		Socket s = null;

		SJStartDelegation sjsd = null;

		try
		{
			ss = new ServerSocket(port);

			sjsd = (SJStartDelegation) readObjectNoCNFEWithDelCheck();

			//SJRuntimeUtils.debugPrintln("Start delegation report: " + sjsd.socketParameters());

			writeByte(SJ_INT);
			writeIntAndFlush(port);

			//SJRuntimeUtils.debugPrintln("Waiting for reconnection on: " + port);

			s = ss.accept(); // Later, allow timeouts.

			//SJRuntimeUtils.debugPrintln("Accepted reconnection on " + port  + " from: " + s.getInetAddress().getHostName() + ":"+ s.getPort());

			s.setTcpNoDelay(true);
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
		finally
		{
			try
			{
				SJRuntimeUtils.closeServerSocket(ss);
			}
			catch (IOException ioe) { }
		}

		SJRSocket rsocket = new SJRSocket(decodeSJType(protocol), protocol, s);

		rsocket.setSocketParameters(sjsd.socketParameters());

		rsocket.acceptIOStreams(); // "Server" side (connection renegotiation does `connect').
		rsocket.isActive(true);
		rsocket.setOriginalRequestor(sjsd.originalRequestor());

		byte simdel = rsocket.readByte();

		//SJRuntimeUtils.debugPrintln("Simultaneous delegation flag: " + simdel);

		if (simdel == SIM_DEL)
		{
			SJDelegationSignal sjds = (SJDelegationSignal) rsocket.readObjectNoCNFE();

			try
			{
				for (byte flag = rsocket.readByte(); flag != SJ_END_OF_RESEND; flag = rsocket.readByte())
				{
					Object obj = null;

					if (flag == SJ_SERIALIZED)
					{
						obj = rsocket.readSerialized(); // The lost message has been deserialized.
					}
					else if (flag == SJ_BOOLEAN)
					{
						obj = new Boolean(rsocket.readBoolean());
					}
					else if (flag == SJ_INT)
					{
						obj = new Integer(rsocket.readInt());
					}
					else if (flag == SJ_BYTE)
					{
						obj = new Byte(rsocket.readByte());
					}
					else
					{
						throw new SJIOException("Unrecognised message flag: " + flag);
					}

					rsocket.lostMessages.add(obj);

					//SJRuntimeUtils.debugPrintln("Buffered lost message: " + obj);
				}
			}
			catch (ClassNotFoundException cnfe)
			{
				throw new SJIOException(cnfe);
			}

			rsocket.shutdown(); // No state reset.

			String targetAddr = sjds.hostAddress();
			int targetPort = sjds.port(); // Don't need the session type component of the delegation signal.

			//SJRuntimeUtils.debugPrintln("Reconnecting for simulatenous delegation to: " + targetAddr + ":" + targetPort);

			rsocket.connect(new InetSocketAddress(targetAddr, targetPort));

			rsocket.writeByteAndFlush(SINGLE_DEL);
		}
		else //if (simdel.equals(SINGLE_DEL))
		{
			// Nothing special to do.
		}

		rsocket.state.open();

		state.receiveSession(decodeSJType(encoded));

		//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());

		return rsocket;
	}

	private final void reconnectSingleDelegation(SJDelegationSignal theirds) throws SJIOException
	{
		String targetAddr = theirds.hostAddress();
		int targetPort = theirds.port();
		SJType targetState = decodeSJType(theirds.targetState());

		//SJRuntimeUtils.debugPrintln("Single delegation signal report: " + targetAddr + ":" + targetPort + ", " + targetState);

		try
		{
			writeByteAndFlush(SJ_DEL_ACK); // Will fail for delegation cases 2 and 3 (session-sender has already shutdown the session being delegated)...
		}
		catch (SJIOException ioe) { } //...so exception suppressed.
		finally
		{
			try
			{
				shutdown();
			}
			catch (SJIOException sjioe) { }
		}

		SJType currentState = currentState();

		//SJRuntimeUtils.debugPrintln("Current state: " + currentState);

		int resendNumber = calculateResendNumber(currentState, targetState);

		//SJRuntimeUtils.debugPrintln("Session units to resend: " + resendNumber);

		connect(new InetSocketAddress(targetAddr, targetPort));

		writeByteAndFlush(SINGLE_DEL);

		if (resendNumber > 0)
		{
			printSentMessageCache();
			resendLostMessages(resendNumber);
		}
	}

	private final void reconnectSimultaneousDelegation(SJDelegationSignal theirds, SJDelegationSignal ourds) throws SJIOException
	{
		String targetAddr = theirds.hostAddress();
		int targetPort = theirds.port();
		SJType targetState = decodeSJType(theirds.targetState());

		//SJRuntimeUtils.debugPrintln("Simultaneous delegation signal report: " + targetAddr + ":" + targetPort + ", " + targetState);

		try
		{
			//writeByteAndFlush(SJ_DEL_ACK); // Another ACK not necessary, the delegation signal is enough.
		}
		finally
		{
			try
			{
				shutdown();
			}
			catch (SJIOException sjioe) { }
		}

		SJType currentState = currentState();

		//SJRuntimeUtils.debugPrintln("Current state: " + currentState);

		int resendNumber = calculateResendNumber(currentState, targetState);

		//SJRuntimeUtils.debugPrintln("Session units to resend: " + resendNumber);

		if (resendNumber < 0)
		{
			// The session peer needs to resend some lost messages, we don't do anything.
		}
		else if (resendNumber == 0)
		{
			if (isOriginalRequestor())
			{
				connect(new InetSocketAddress(targetAddr, targetPort));

				writeByte(SIM_DEL);
				writeObject(ourds);

				writeByteAndFlush(SJ_END_OF_RESEND);
			}
			else
			{
				// The "original acceptor" doesn't need to do anything else.
			}
		}
		else //if (resendNumber > 0)
		{
			connect(new InetSocketAddress(targetAddr, targetPort));

			writeByte(SIM_DEL);
			writeObject(ourds);

			printSentMessageCache();
			resendLostMessages(resendNumber);

			writeByteAndFlush(SJ_END_OF_RESEND);
		}
	}

	private final void resendLostMessages(int resendNumber) throws SJIOException // Could move these calculations and the resend operations into a SentMessageCache object, or combine with the socket state?
	{
		int cacheSize = sentMessageCache.size();
		int resendPos = (cacheSize == 0) ? 0 :cacheSize - 1;

		while (resendPos > 0 && sentMessageCache.get(resendPos) != CACHE_MARKER)
		{
			resendPos--;
		}

		if (resendPos > 0) // Why do this? Instead take marker into account when going backwards for each resendNumber. Would be easier to think about if we're always pointing to a marker?
		{
			resendPos--;
		}

		while (resendPos > 0 && resendNumber > 0)
		{
			if (sentMessageCache.get(resendPos) == CACHE_MARKER)
			{
				resendNumber--;

				if (resendNumber > 0) // Could instead "add back on" after the loop.
				{
					resendPos--;
				}
			}
			else
			{
				resendPos--;
			}
		}

		//SJRuntimeUtils.debugPrintln("Resending from cache position: " + resendPos);

		while (resendPos < cacheSize)
		{
			SJRSerializedMessage sermsg = sentMessageCache.get(resendPos++);

			if (!(sermsg == CACHE_MARKER))
			{
				if (sermsg.isPrimitive())
				{
					if (sermsg.isBoolean())
					{
						//SJRuntimeUtils.debugPrintln("Resending: " + sermsg.booleanMessage());

						writeByte(SJ_BOOLEAN); //writeSerialized protocol (not a `send', no state update).
						writeBoolean(sermsg.booleanMessage());
					}
					else if (sermsg.isInt())
					{
						//SJRuntimeUtils.debugPrintln("Resending: " + sermsg.intMessage());

						writeByte(SJ_INT);
						writeIntAndFlush(sermsg.intMessage());
					}
					else if (sermsg.isByte())
					{
						//SJRuntimeUtils.debugPrintln("Resending: " + sermsg.byteMessage());

						writeByte(SJ_BYTE);
						writeByteAndFlush(sermsg.byteMessage());
					}
				}
				else
				{
					try
					{
						//SJRuntimeUtils.debugPrintln("Resending: " + deserializeObject(sermsg.objectMessage()));
					}
					catch (Exception x)
					{
						throw new SJIOException(x);
					}

					writeByte(SJ_SERIALIZED);
					writeSerialized(sermsg.objectMessage()); // The other side will read and deserialize the object implicitly, whether it will be received "normally", or as a buffered lost message.
				}
			}
		}
	}

	private static final int calculateResendNumber(SJType current, SJType target)
	{
		int resendNumber = 0;

		while (current != null)
		{
			if (current instanceof SJOutbranchType) // Can optimise: if this is not the last element in session type and target isn't null, then don't need to check, i.e. the opposing element will be the exact dual.
			{
				SJOutbranchType obt = (SJOutbranchType) current;
				SJLabel lab = obt.labelSet().iterator().next(); // Socket state only records the selected label. // For iteration, only the final selection is recorded, but doesn't matter inside an iteration.

				resendNumber += calculateResendNumber(obt.branchCase(lab), (target == null) ? null : ((SJInbranchType) target).branchCase(lab));
			}

			if (target == null)
			{
				resendNumber++; // Sends the label, in the outbranch case.
			}
			else
			{
				target = target.child();
			}

			current = current.child(); // Steps over whole iterations.
		}

		// If target is terminal (then current must also be terminal) - session already completed, no need to resend any messages (or even delegate!).
		return (current == null && target != null) ? -1 : resendNumber; // Exact "negative difference" doesn't matter.
	}

	public final void sendServerAddress(SJServerAddress sjsa) throws SJIOException
	{
		try // Maybe this should be factored out, also used by send and outlabel.
		{
			byte[] serobj = serializeObject(sjsa.targetAddress()); // Optimise for byte arrays?

			writeByte(SJ_SERIALIZED); // Make a SJ_CHANNEL flag?
			writeSerialized(serobj);

			if (RUNTIME_STATE)
			{
				pushSentMessage(new SJRSerializedMessage(serobj));

				state.sendChannel(decodeSJType(sjsa.getProtocol().encodedSJType()));

				//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
			}
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	public final SJServerAddress receiveServerAddress(String encoded) throws SJIOException
	{
		InetSocketAddress isa = (InetSocketAddress) readObjectNoCNFEWithDelCheck();
		SJServerAddress channel = SJServerAddress.create(new SJProtocol(encoded), isa);

		if (RUNTIME_STATE)
		{
			state.receiveChannel(decodeSJType(encoded));

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return channel;
	}

	private final class closeProtocol implements Runnable
	{
		public final void run()
		{
			try
			{
				while (true) // In case of re-delegations.
				{
					writeByteAndFlush(SJ_FIN);

					if (lostMessages.size() == 0)
					{
						byte flag = readByte();

						if (flag == SJ_FIN)
						{
							break;
						}
						else if (flag == SJ_DEL_SIGNAL)
						{
							reconnectSingleDelegation((SJDelegationSignal) readObjectNoCNFE()); // Sends an ACK, but this action will fail (exception suppressed) since session-sender closes the delegated session immediately after getting our FIN.
						}
						else if (flag == SJ_SERIALIZED || flag == SJ_START_DEL) // Session has failed, discard these messages.
						{
							try
							{
								readSerialized();
							}
							catch (ClassNotFoundException cnfe)
							{
								// We don't care if we can't deserialize, but does suppressing the exception work?
							}
						}
						else // Primitives.
						{
							if (flag == SJ_BOOLEAN)
							{
								readBoolean();
							}
							else if (flag == SJ_INT)
							{
								readInt();
							}
							else if (flag == SJ_BYTE)
							{
								readByte();
							}
							else
							{
								// Bad message flag.
							}
						}
					}
					else
					{
						try
						{
							if (readLostMessage() instanceof SJFIN) // Else oiscard.
							{
								break;//redo = false;
							}
						}
						catch (ClassNotFoundException cnfe)
						{
							// Don't care if we can't deserialize lost messages if session has failed.
						}
					}
				}
			}
			catch (SJIOException ioe)
			{
				//ioe.printStackTrace(); // Suppress?
			}
			finally
			{
				try
				{
					//SJRuntimeUtils.debugPrintln("Closing socket (" + getLocalAddress().getPort() + ")" + (RUNTIME_STATE ? (": " + state.currentState()) : ""));

					if (RUNTIME_STATE)
					{
						state.close();
					}

					shutdown();
				}
				catch (SJIOException sjioe) { }
			}
		}
	}

	protected final void close() // Almost the same as in SJFSocket.
	{
		if (isActive())
		{
			isActive(false);

			spawnCloseThread();
		}
	}

	private final void spawnCloseThread()
	{
		//SJRuntimeUtils.debugPrintln("Spawning close thread for (" + getLocalAddress().getPort() + ")" + (RUNTIME_STATE ? (": " + state.currentState()) : ""));

		new Thread(new closeProtocol()).start();
	}

	private final void shutdown() throws SJIOException
	{
		try
		{
			SJRuntimeUtils.closeStream(ois);
			SJRuntimeUtils.closeStream(oos);
			SJRuntimeUtils.closeSocket(getSocket());
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	public final void send(Object obj) throws SJIOException
	{
		try
		{
			byte[] serobj = serializeObject(obj); // Optimise for byte arrays?

			writeByte(SJ_SERIALIZED);
			writeSerialized(serobj);

			if (RUNTIME_STATE)
			{
				pushSentMessage(new SJRSerializedMessage(serobj));

				if (!state.inIterationContext()) // Can't be moved into writeCachedObject. // Why?
				{
					pushSentMessage(CACHE_MARKER); // Could move this logic into a SentMessageCache object (would need access to socket state - or could make the cache part of the socket state?)
				}

				state.send(obj);

				//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
			}
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	public final void sendBoolean(boolean v) throws SJIOException
	{
		writeByte(SJ_BOOLEAN);
		writeBoolean(v);

		if (RUNTIME_STATE)
		{
			pushSentMessage(new SJRSerializedMessage(v));

			if (!state.inIterationContext())
			{
				pushSentMessage(CACHE_MARKER);
			}

			state.sendBoolean(v);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}
	}

	public final void sendInt(int v) throws SJIOException
	{
		writeByte(SJ_INT);
		writeIntAndFlush(v);

		if (RUNTIME_STATE)
		{
			pushSentMessage(new SJRSerializedMessage(v));

			if (!state.inIterationContext())
			{
				pushSentMessage(CACHE_MARKER);
			}

			state.sendInt(v);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}
	}

	public final void sendByte(byte v) throws SJIOException
	{
		writeByte(SJ_BYTE);
		writeByteAndFlush(v);

		if (RUNTIME_STATE)
		{
			pushSentMessage(new SJRSerializedMessage(v));

			if (!state.inIterationContext())
			{
				pushSentMessage(CACHE_MARKER);
			}

			state.sendByte(v);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}
	}

	public final Object receive() throws SJIOException, ClassNotFoundException
	{
		Object obj = readObjectWithDelegationCheck(); // Clears send message cache.

		if (RUNTIME_STATE)
		{
			state.receive(obj);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return obj;
	}

	public final boolean receiveBoolean() throws SJIOException
	{
		boolean v = readBooleanWithDelegationCheck();

		if (RUNTIME_STATE)
		{
			state.receiveBoolean(v);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return v;
	}

	public final int receiveInt() throws SJIOException
	{
		int v = readIntWithDelegationCheck();

		if (RUNTIME_STATE)
		{
			state.receiveInt(v);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return v;
	}

	public final byte receiveByte() throws SJIOException
	{
		byte v = readByteWithDelegationCheck();

		if (RUNTIME_STATE)
		{
			state.receiveByte(v);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return v;
	}

	public final String outlabel(String lab) throws SJIOException
	{
		try
		{
			byte[] serlab = serializeObject(lab);

			writeByte(SJ_SERIALIZED); // Make a SJ_LABEL flag?
			writeSerialized(serlab);

			if (RUNTIME_STATE)
			{
				pushSentMessage(new SJRSerializedMessage(serlab));

				if (!state.inIterationContext())
				{
					pushSentMessage(CACHE_MARKER);
				}

				state.outbranch(new SJLabel(lab));

				//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
			}
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}

		return lab;
	}

	public final String inlabel() throws SJIOException
	{
		String lab = (String) readObjectNoCNFEWithDelCheck();

		if (RUNTIME_STATE)
		{
			state.inbranch(new SJLabel(lab));

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return lab;
	}

	public final boolean outsync(boolean bool) throws SJIOException
	{
		writeByte(SJ_BOOLEAN);
		writeBoolean(bool);

		if (RUNTIME_STATE)
		{
			pushSentMessage(new SJRSerializedMessage(bool));

			if (!bool && !state.inIterationContext())
			{
				pushSentMessage(CACHE_MARKER);
			}

			state.outwhile(bool);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return bool;
	}

	public final boolean insync() throws SJIOException
	{
		boolean bool = readBooleanWithDelegationCheck();

		if (RUNTIME_STATE)
		{
			state.inwhile(bool);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return bool;
	}

	public final boolean recursionFalse(String lab) throws SJIOException
	{
		if (RUNTIME_STATE)
		{
			state.recursion(new SJLabel(lab));

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return false;
	}

	public final boolean recursionTrue(String lab) throws SJIOException
	{
		if (RUNTIME_STATE)
		{
			state.recurse(new SJLabel(lab));
		}

		return true;
	}

	private final void writeObject(Object obj) throws SJIOException
	{
		try
		{
			oos.writeObject(obj);
			oos.flush();
			//bos.flush();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final void writeSerialized(byte[] serobj) throws SJIOException
	{
		try
		{
			oos.writeInt(serobj.length); // Alternative is to encapsulate in SJRSerializedMessage (that means primitives would be sent as a single object rather than an int and a primitive).
			oos.write(serobj, 0, serobj.length);
			oos.flush();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final void writeBoolean(boolean b) throws SJIOException
	{
		try
		{
			oos.writeBoolean(b);
			oos.flush();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final void writeByte(byte v) throws SJIOException // Rename writeFlag?
	{
		try
		{
			oos.writeByte(v); // Not flushed.
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final void writeByteAndFlush(byte v) throws SJIOException
	{
		try
		{
			oos.writeByte(v);
			oos.flush();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	/*private final void writeInt(int v) throws SJIOException // Rename write(Serialized)Length?
	{
		try
		{
			oos.writeInt(v); // Not flushed.
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}*/

	private final void writeIntAndFlush(int v) throws SJIOException
	{
		try
		{
			oos.writeInt(v);
			oos.flush();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final Object readObject() throws SJIOException, ClassNotFoundException
	{
		try
		{
			return ois.readObject();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final Object readSerialized() throws SJIOException, ClassNotFoundException
	{
		try
		{
			byte[] serobj = new byte[ois.readInt()];

			ois.readFully(serobj);

			return deserializeObject(serobj);
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final boolean readBoolean() throws SJIOException
	{
		try
		{
			return ois.readBoolean();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final byte readByte() throws SJIOException
	{
		try
		{
			return ois.readByte();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final int readInt() throws SJIOException
	{
		try
		{
			return ois.readInt();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final Object readObjectNoCNFE() throws SJIOException
	{
		try
		{
			return readObject();
		}
		catch (ClassNotFoundException cnfe)
		{
			throw new SJIOException(cnfe);
		}
	}

	private final Object readObjectWithDelegationCheck() throws SJIOException, ClassNotFoundException
	{
		if (lostMessages.size() == 0)
		{
			byte flag = readByte();

			if (flag == SJ_SERIALIZED)
			{
				Object obj = readSerialized();

				sentMessageCache.clear();

				return obj;
			}
			else if (/*flag == SJ_OBJECT || */flag == SJ_START_DEL)
			{
				return readObjectNoCNFE();
			}
			else if (flag == SJ_DEL_SIGNAL)
			{
				reconnectSingleDelegation((SJDelegationSignal) readObjectNoCNFE());

				return readObjectWithDelegationCheck(); // Session could be re-delegated.
			}
			else //if (flag == SJ_FIN)
			{
				throw new SJIOException("Session prematurely closed by peer.");
			}
		}
		else
		{
			return readLostMessage();
		}
	}

	private final Object readLostMessage() throws SJIOException, ClassNotFoundException
	{
		Object obj = obj = lostMessages.remove(0);

		//SJRuntimeUtils.debugPrintln("Loaded lost message: " + obj);

		return obj;
	}

	private final boolean readBooleanWithDelegationCheck() throws SJIOException
	{
		if (lostMessages.size() == 0)
		{
			byte flag = readByte();

			if (flag == SJ_BOOLEAN)
			{
				boolean b = readBoolean();

				sentMessageCache.clear();

				return b;
			}
			else if (flag == SJ_DEL_SIGNAL)
			{
				reconnectSingleDelegation((SJDelegationSignal) readObjectNoCNFE()); // FIN possible?

				return readBooleanWithDelegationCheck(); // Could be re-delegation.
			}
			else
			{
				throw new SJIOException("Session prematurely closed by peer.");
			}
		}
		else
		{
			try
			{
				return ((Boolean) readLostMessage()).booleanValue();
			}
			catch (ClassNotFoundException cnfe)
			{
				throw new SJIOException(cnfe);
			}
		}
	}

	private final int readIntWithDelegationCheck() throws SJIOException
	{
		if (lostMessages.size() == 0)
		{
			byte flag = readByte();

			if (flag == SJ_INT)
			{
				int i = readInt();

				sentMessageCache.clear();

				return i;
			}
			else if (flag == SJ_DEL_SIGNAL)
			{
				reconnectSingleDelegation((SJDelegationSignal) readObjectNoCNFE());

				return readIntWithDelegationCheck(); // Could be re-delegation.
			}
			else
			{
				throw new SJIOException("Session prematurely closed by peer.");
			}
		}
		else
		{
			try
			{
				return ((Integer) readLostMessage()).intValue();
			}
			catch (ClassNotFoundException cnfe)
			{
				throw new SJIOException(cnfe);
			}
		}
	}

	private final byte readByteWithDelegationCheck() throws SJIOException
	{
		if (lostMessages.size() == 0)
		{
			byte flag = readByte();

			if (flag == SJ_BYTE)
			{
				byte i = readByte();

				sentMessageCache.clear();

				return i;
			}
			else if (flag == SJ_DEL_SIGNAL)
			{
				reconnectSingleDelegation((SJDelegationSignal) readObjectNoCNFE());

				return readByteWithDelegationCheck(); // Could be re-delegation.
			}
			else
			{
				throw new SJIOException("Session prematurely closed by peer.");
			}
		}
		else
		{
			try
			{
				return ((Byte) readLostMessage()).byteValue();
			}
			catch (ClassNotFoundException cnfe)
			{
				throw new SJIOException(cnfe);
			}
		}
	}

	private final Object readObjectNoCNFEWithDelCheck() throws SJIOException
	{
		try
		{
			return readObjectWithDelegationCheck();
		}
		catch (ClassNotFoundException cnfe)
		{
			throw new SJIOException(cnfe);
		}
	}

	private ByteArrayOutputStream baos = new ByteArrayOutputStream(INIT_SER_BUF_SIZE); // If the byte array output stream is reused (as a field), then it will (roughly) be the size of the largest message sent so far for the duration the session. Will this affect the performance of sending smaller messages (byte array operations)? On the other hand, creating a new byte array output stream for each message send will be affected by the default buffer size: large messages will need data to be copied to successively bigger arrays, but less penalty for smaller messages?

	private final byte[] serializeObject(Object obj) throws IOException // static?
	{
		//BufferedOutputStream bos = null;
		ObjectOutputStream os = null; // Reusing these stream objects (as fields) doesn't work.

		try
		{
			//baos = new ByteArrayOutputStream(INIT_SER_BUF_SIZE); // Factor out constant.
			//bos = new BufferedOutputStream(baos); // Buffer size?
			os = new ObjectOutputStream(baos);
			//os = new JBossObjectOutputStream(baos);

			os.writeObject(obj);
			os.flush();

			return baos.toByteArray();
		}
		finally
		{
			try
			{
				SJRuntimeUtils.closeStream(os);
				//SJRuntimeUtils.closeStream(bos);

				baos.reset(); // Closing a byte array stream does nothing.
			}
			catch (IOException ioe)
			{
				//throw new RuntimeException(ioe);
			}
		}
	}

	private final Object deserializeObject(byte[] byteArray) throws IOException, ClassNotFoundException
	{
		ByteArrayInputStream bais = null;
		ObjectInputStream is = null;

		try
		{
			bais = new ByteArrayInputStream(byteArray);

			if (!remoteClassLoadingEnabled())
			{
				is = new ObjectInputStream(bais);
				//is = new JBossObjectInputStream(bis);
			}
			else
			{
				is = new SJObjectInputStream(bais, getRemoteClassLoader());
			}

			return is.readObject();
		}
		finally
		{
			try
			{
				SJRuntimeUtils.closeStream(is);
			}
			catch (IOException ioe)
			{
				//throw new RuntimeException(ioe);
			}
		}
	}

	protected final void init() throws SJIOException, SJIncompatibleSessionException // Can be moved to SJSocket? (Only if the required read and write methods are moved too.) // Maybe eager class loading and verification should also be done after a session is delegated.
	{
		SJType protocol = protocolType();

		try
		{
			boolean ourecd = eagerClassDownloadingEnabled();
			boolean ourecv = eagerClassVerificationEnabled();

			oos.writeBoolean(ourecd);
			oos.writeBoolean(ourecv);
			oos.flush(); // Any gains by manually flushing both booleans together?

			boolean theirecd = ois.readBoolean();
			boolean theirecv = ois.readBoolean();

			if (theirecd || theirecv)
			{
				//try // Collapsing try scopes will be faster?
				{
					writeObject(SJRuntimeUtils.getClassSVUIDs(protocol)); // Refactor and optimise for class verification. // Need to modify to include interfaces? // Could cache class SVUIDs map for future use.
				}
				/*catch (ClassNotFoundException cnfe)
				{
					throw new SJIOException(cnfe);
				}*/
			}

			if (ourecd || ourecv)
			{
				Map<String, Long> theirClassSVUIDs = null;

				theirClassSVUIDs = (Map<String, Long>) readObjectNoCNFE(); // No need to check for delegation.

				if (ourecd)
				{
					eagerClassDownload(protocol, theirClassSVUIDs);
				}

				if (ourecv)
				{
					checkClasses(theirClassSVUIDs);
				}
			}

			if (theirecv)
			{
				//verifyClasses(ourecv, theirecv, protocol, theirClassSVUIDs);

				Object ack = readObjectNoCNFE(); // Ack sent by checkClasses.

				if (!((String) ack).equals(CLASSES_VERIFIED))
				{
					throw new SJIncompatibleSessionException("IncompatibleClass: " + ack);
				}
			}

			if (CHECK_DUALITY)
			{
				checkDuality(protocol);
			}
		}
		catch (ClassNotFoundException cnfe)
		{
			throw new SJIOException(cnfe);
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
		catch (SJException x)
		{
			shutdown();

			if (x instanceof SJIncompatibleSessionException)
			{
				throw (SJIncompatibleSessionException) x;
			}
			else //if (x instanceof SJIOException)
			{
				throw (SJIOException) x;
			}
		}
	}

	private final void eagerClassDownload(SJType protocolType, Map<String, Long> theirMap) throws SJIOException
	{
		for (Entry<String, Long> e : theirMap.entrySet())
		{
			String className = e.getKey();
			Class c = null;

			try
			{
				c = Class.forName(className); // Any faster way to check this?
			}
			catch(ClassNotFoundException cnfe1)
			{
				try
				{
					c = getRemoteClassLoader().loadClass(className); // If we have just downloaded this class, then we implicitly do not need to verify it.
				}
				catch (ClassNotFoundException cnfe2)
				{
					throw new SJIOException(cnfe2);
				}

				//SJRuntimeUtils.debugPrintln("Eagerly loaded: " + c);
			}
		}
	}

	/*private final void verifyClasses(boolean ourecv, boolean theirecv, SJType protocolType, Map<String, Long> theirMap) throws SJIOException, SJIncompatibleSessionException
	{

	}*/

	private final void checkClasses(Map<String, Long> theirs) throws SJIOException,  SJIncompatibleSessionException
	{
		for (Entry<String, Long> e : theirs.entrySet())
		{
			String className = e.getKey();
			Class c = null;

			try
			{
				if (remoteClassLoadingEnabled()) // Codebase is set (not null), already checked.
				{
					c = Class.forName(className, true, getRemoteClassLoader());
				}
				else
				{
					c = Class.forName(className);
				}
			}
			catch(ClassNotFoundException cnfe) // Class not in codebase or remote class loading not enabled.
			{
				throw new SJIncompatibleSessionException("Could not verify class: " + className); // Actually can't get here: would have already failed trying to decode the peer's session type (all classes are needed to decode).
			}

			ObjectStreamClass osc = ObjectStreamClass.lookup(c);

			if (!(e.getValue().longValue() == osc.getSerialVersionUID()))
			{
				writeObject(className); // Report the bad class.

				throw new SJIncompatibleSessionException("Incompatible class: " + className);
			}

			//SJRuntimeUtils.debugPrintln("Eagerly checked class: " + className);
		}

		writeObject(CLASSES_VERIFIED);
	}

	private final void checkDuality(SJType ours) throws SJIOException, SJIncompatibleSessionException
	{
		writeObject(getChannel().getProtocol().encodedSJType());

		SJType theirs = decodeSJType((String) readObjectNoCNFE()); // Will throw NullPointerException if any classes are not available (and remote class loading oisabled or class not available from repository). // Possible to get premature SJ_FIN here? (Only if thread is interrupted?)

		if (!ours.treeDualtype(theirs))
		{
			throw new SJIncompatibleSessionException("Our session type \"" + ours + "\" incompatible with requested session type \"" + theirs + "\".");
		}

		//SJRuntimeUtils.debugPrintln("Session initiated: our type `" + ours + "', their type `" + theirs + "'");
	}

	public final SJType protocolType()
	{
		return protocolType;
	}

	public final SJType currentState()
	{
		return RUNTIME_STATE ? state.currentState() : null;
	}

	private final SJType expectedType()
	{
		return RUNTIME_STATE ? state.expectedType() : null;
	}

	protected final void pushSentMessage(SJRSerializedMessage sermsg) // Booleans need to be encapsulated properly, not just boxed (otherwise don't know if we should be resending a primitive or boxed - in the current resending algorithm, since it is done by resend position and not from the types directly).
	{
		sentMessageCache.push(sermsg);
		printSentMessageCache();
	}

	private final void printSentMessageCache()
	{
		String m = "Sent message cache: ";

		try
		{
			for (int size = sentMessageCache.size(), i = 0; i < size; i++)
			{
				SJRSerializedMessage sermsg = sentMessageCache.get(i);

				if (sermsg == CACHE_MARKER)
				{
					if (m.endsWith(", "))
					{
						m = m.substring(0, m.length() - 2);
					}

					m = m + " | ";
				}
				else
				{
					String msg = null;

					if (sermsg.isPrimitive())
					{
						if (sermsg.isBoolean())
						{
							msg = Boolean.toString(sermsg.booleanMessage());
						}
						else if (sermsg.isInt())
						{
							msg = Integer.toString(sermsg.intMessage());
						}
					}
					else
					{
						msg = deserializeObject(sermsg.objectMessage()).toString();
					}

					m = m + msg;

					if (i < size - 1)
					{
						m = m + ", ";
					}
				}
			}
		}
		catch (Exception x)
		{
			throw new RuntimeException(x);
		}

		//SJRuntimeUtils.debugPrintln(m);
	}
}
