/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.runtime.net;

import java.net.*;

import polyglot.types.SemanticException;

import sj.ExtensionInfo;
import sj.types.SJTypeSystem;
import sj.types.typeobjects.sesstypes.SJType;
import sj.util.SJTypeEncoder;

import sj.runtime.*;
import sj.runtime.util.*;

public abstract class SJRemoteClassHandler // Unify with SJRemoteClassLoader? Then all sockets and server sockets would become remote class loaders. // Has also taken on the runtime type system elements common to both to support the type encoding/decoding operations that may trigger the remote class loader.
{
	private	static final ExtensionInfo extInfo;
	private static final SJTypeSystem sjts;
	private static final SJClassResolver sjcr;
	private static final SJTypeEncoder sjte;
	//private static Object sjteLock = new Object(); // Can we just synchronise on sjte directly?

	private SJRemoteClassLoader sjrcl; // Might not be an adequate solution: will have separate name spaces for actual class loading (remote class loader per socket), but a single static resolver might not be able to have different classes with the same name - a problem for class verification and type decoding.

	private SJSocketParameters params = new SJSocketParameters(); // Allow multiple codebases?

	static
	{
		try
		{
			extInfo = new ExtensionInfo();
			sjts = (SJTypeSystem) extInfo.typeSystem();
			sjcr = new SJClassResolver(sjts, extInfo);
			sjts.initialize(sjcr, extInfo);
			sjte = new SJTypeEncoder(sjts);
		}
		catch (SemanticException se)
		{
			throw new RuntimeException(se);
		}
	}

	protected final SJTypeSystem sjTypeSystem()
	{
		return sjts;
	}

	public final void checkClassLoadingParameters() throws SJIOException
	{
		if (remoteClassLoadingEnabled())
		{
			if (getCodebase() == null)
			{
				throw new SJIOException("eagerClassDownloading enabled but no codebase set.");
			}
		}
		else
		{
			if (eagerClassDownloadingEnabled())
			{
				throw new SJIOException("Cannot perform eager class downloading when remoteClassLoading is disabled.");
			}
			/*else if (eagerClassVerificationEnabled())
			{
				// Ok to try eager class verification without remote class loading (if it is enabled, however, the above check makes sure the codebase is also set).
			}*/
		}
	}

	public final SJSocketParameters getSocketParameters()
	{
		return params;
	}

	protected final boolean isOriginalRequestor() throws SJIOException // Although this doesn't have much to do with remote class handling, it's part of the parameters object.
	{
		return params.isOriginalRequestor();
	}

	public final URL getCodebase()
	{
		return params.getCodebase();
	}

	public final boolean eagerClassDownloadingEnabled()
	{
		return params.eagerClassDownloadingEnabled();
	}

	public final boolean remoteClassLoadingEnabled()
	{
		return params.remoteClassLoadingEnabled();
	}

	public final boolean eagerClassVerificationEnabled()
	{
		return params.eagerClassVerificationEnabled();
	}

	// The setters are overridden by SJSocket. (Bad design?)
	public void setSocketParameters(SJSocketParameters params) throws SJIOException
	{
		this.params = params;
	}

	protected final void setOriginalRequestor(boolean origreq) throws SJIOException
	{
		params.setOriginalRequestor(origreq);
	}

	public void setRemoteClassLoading(boolean enabled) throws SJIOException
	{
		params.setRemoteClassLoading(enabled);
	}

	public void setEagerClassDownloading(boolean eager) throws SJIOException
	{
		params.setEagerClassDownloading(eager);
	}

	public void setEagerClassVerification(boolean eager) throws SJIOException
	{
		params.setEagerClassVerification(eager);
	}

	public void setCodebase(URL codebase) throws SJIOException
	{
		params.setCodebase(codebase);
	}

	public void setCodebase(String codebase) throws SJIOException
	{
		try
		{
			setCodebase(new URL(codebase));
		}
		catch (MalformedURLException mue)
		{
			throw new SJIOException(mue);
		}
	}

	protected final String encodeSJType(SJType sjType) throws SJIOException
	{
		synchronized (SJRemoteClassHandler.sjte) // In case multiple threads (different sockets) want to decode a protocol at the same time: only one class resolver, so need to use the corresponding class loader (for the correct class loading context) for each separate request.
		{
			SJRemoteClassHandler.sjcr.setRemoteClassLoader(this.sjrcl); // We only need to manually set the remote class loader on the class resolver for protocol encoding and decoding (that's the only purpose for the class resolver), not to actually receive a message (class loader used directly by SJObjectInputStream - and if classes were eagerly loaded, then the object input stream can just get the cached binary).

			try
			{
				return SJRemoteClassHandler.sjte.encode(sjType);
			}
			catch (SemanticException se)
			{
				throw new RuntimeException(se);
			}
			finally
			{
				//SJSocket.sjte.notifyAll();
			}
		}
	}

	public final synchronized SJType decodeSJType(String encoded) throws SJIOException // Hacked access for benchmarking.
	{
		//synchronized (SJRemoteClassHandler.sjte) // Just declare the method as synchronized?
		{
			SJRemoteClassHandler.sjcr.setRemoteClassLoader(this.sjrcl); // Remote class loader needed in case peer's session type contains classes not locally available. If remote class loading is enabled (and codebase is set), we can try to download the class; otherwise decoding fails. Consequently, class downloading due to eager verification is currently never needed. // Remote class loader is null for protocol decoding for server create.

			try
			{
				return SJRemoteClassHandler.sjte.decode(encoded);
			}
			catch (SemanticException se)
			{
				throw new RuntimeException(se);
			}
			finally
			{
				//SJSocket.sjte.notifyAll();
			}
		}
	}

	protected final SJType decodeSJType(SJProtocol protocol) // Protocols have been checked by compiler.
	{
		try
		{
			SJType sjType = decodeSJType(protocol.encodedSJType());

			//SJRuntimeUtils.debugPrintln("Decoded protocol: " + sjType);

			return sjType;
		}
		catch (SJIOException ioe)
		{
			throw new RuntimeException(ioe);
		}
	}

	protected final ExtensionInfo extensionInfo()
	{
		return extInfo;
	}

	protected final SJRemoteClassLoader getRemoteClassLoader()
	{
		return sjrcl;
	}

	protected final void setRemoteClassLoader(SJRemoteClassLoader sjrcl)
	{
		this.sjrcl = sjrcl;
	}
}
