/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.runtime.net;

import java.net.*;

import sj.SJConstants;
import sj.types.typeobjects.sesstypes.SJType;

import sj.runtime.*;

public abstract class SJSocket extends SJRemoteClassHandler
{
	protected static final int UNBOUND_PORT_VALUE = -1;

	public static final boolean TCP_NO_DELAY = true; // Hacked access for alt subpackage.
	protected static final boolean IS_ORIGINAL_REQUESTOR = true;
	protected static final boolean NOT_ORIGINAL_REQUESTOR = false;
	protected static final String CLASSES_VERIFIED = "CLASSES_VERIFIED"; // Maybe faster to send e.g. an integer value.

	/*protected static final int SJ_OBJECT = 0;
	protected static final int SJ_BOOLEAN = 1;
	protected static final int SJ_BYTE = 2;
	protected static final int SJ_CHAR = 3;
	protected static final int SJ_INT = 4;*/

	private int spawnCount = SJConstants.SJ_SPAWN_COUNT_INIT;
	private boolean isActive = false;

	private SJServerAddress channel;
	private Socket socket;

	abstract public void request() throws UnknownHostException, SJIOException, SJIncompatibleSessionException;

	abstract public void send(Object obj) throws SJIOException;
	abstract public void sendBoolean(boolean v) throws SJIOException; // Redundant.
	abstract public void sendInt(int v) throws SJIOException;
	abstract public void sendByte(byte v) throws SJIOException;
	abstract public void sendSession(SJSocket s, String encoded) throws SJIOException; // Take the encoded session type instead? Any advantage?
	abstract public void sendServerAddress(SJServerAddress sjsa) throws SJIOException; // Encode the session type as an argument? // No point? The sjsa already (indirectly) contains the encoded session type (from user implementation, not compiler inferred, but it's static information anyway).

	//abstract public void sendChar(char v) throws SJIOException;
	/*abstract public void sendShort(short v) throws SJIOException;
	abstract public void sendLong(long v) throws SJIOException;
	abstract public void sendFloat(float v) throws SJIOException;
	abstract public void sendDouble(double v) throws SJIOException;*/

	abstract public Object receive() throws SJIOException, ClassNotFoundException;
	abstract public boolean receiveBoolean() throws SJIOException; // Necessary.
	abstract public int receiveInt() throws SJIOException;
	abstract public byte receiveByte() throws SJIOException;
	abstract public SJSocket receiveSession(String encoded) throws SJIOException;
	abstract public SJServerAddress receiveServerAddress(String encoded) throws SJIOException;

	//abstract public char receiveChar() throws SJIOException;
	/*abstract public long receiveLong() throws SJIOException;
	abstract public short receiveShort() throws SJIOException;
	abstract public double receiveDouble() throws SJIOException;
	abstract public float receiveFloat() throws SJIOException;*/

	abstract public String outlabel(String lab) throws SJIOException; // Can make enter and leave methods for the structural operations as targets for compiler translation. Those (enter) methods should call these socket operations. (Session state is currently tracked through these lower level operations.)
	abstract public String inlabel() throws SJIOException;
	abstract public boolean outsync(boolean bool) throws SJIOException;
	abstract public boolean insync() throws SJIOException;
	abstract public boolean recursionFalse(String lab) throws SJIOException; // Exception unnecessary. Also, implementation of this method could be declared here?
	abstract public boolean recursionTrue(String lab) throws SJIOException;

	//abstract public void suspend();
	//abstract public void resume() throws SJIOException;
	//abstract public void resume(int millis) throws SJIOException;

	abstract protected void close() /*throws SJIOException*/; // Does it need to be public?

	public final void close(int spawnCount)
	{
		if (spawnCount == SJConstants.SJ_SPAWN_COUNT_CLOSE || spawnCount == getSpawnCount()) // Can't be greater than.
		{
			close();
		}
	}

	abstract public SJType protocolType();
	abstract public SJType currentState();

	public final int getSpawnCount()
	{
		return spawnCount;
	}

	public final void incrementSpawnCount()
	{
		spawnCount++;
	}

	public final boolean isActive()
	{
		return isActive;
	}

	public final boolean isConnected() // Received "case 2" delegation sockets are active but not connected.
	{
		return getSocket().isConnected();
	}

	public final void isActive(boolean isActive) // Hacked access for alt subpackage.
	{
		this.isActive = isActive;
	}

	protected final SJServerAddress getChannel()
	{
		return channel;
	}

	protected final Socket getSocket()
	{
		return socket;
	}

	protected final void setChannel(SJServerAddress channel)
	{
		this.channel = channel;
	}

	protected final void setSocket(Socket socket)
	{
		this.socket = socket;
	}

	public final InetSocketAddress getLocalAddress()
	{
		return (InetSocketAddress) getSocket().getLocalSocketAddress();
	}

	public final InetSocketAddress getTargetAddress()
	{
		return (InetSocketAddress) getSocket().getRemoteSocketAddress();
	}

	public final void setSocketParameters(SJSocketParameters params) throws SJIOException
	{
		if (isActive())
		{
			throw new SJIOException("Cannot set socket parameters after session established.");
		}

		super.setSocketParameters(params);
	}

	public final void setRemoteClassLoading(boolean enabled) throws SJIOException
	{
		if (isActive())
		{
			throw new SJIOException("Cannot set socket parameter after session established.");
		}

		super.setRemoteClassLoading(enabled);
	}

	public final void setEagerClassDownloading(boolean eager) throws SJIOException
	{
		if (isActive())
		{
			throw new SJIOException("Cannot set socket parameter after session established.");
		}

		super.setEagerClassDownloading(eager);
	}

	public final void setEagerClassVerification(boolean eager) throws SJIOException
	{
		if (isActive())
		{
			throw new SJIOException("Cannot set socket parameter after session established.");
		}

		super.setEagerClassVerification(eager);
	}

	public final void setCodebase(URL codebase) throws SJIOException
	{
		if (isActive())
		{
			throw new SJIOException("Cannot set socket parameter after session established.");
		}

		super.setCodebase(codebase);
	}


	public final void setCodebase(String codebase) throws SJIOException
	{
		if (isActive())
		{
			throw new SJIOException("Cannot set socket parameter after session established.");
		}

		super.setCodebase(codebase);
	}

	// HACK: to satisfy pre-translation type-checking.
	public final void send(boolean v) throws SJIOException // No receive equivalent - return type overloading not possible.
	{
		throw new RuntimeException("Shouldn't get in here: " + v);
	}

	public final void send(int v) throws SJIOException
	{
		throw new RuntimeException("Shouldn't get in here: " + v);
	}

	public final void sendSession(SJSocket s) throws SJIOException
	{
		throw new RuntimeException("Shouldn't get in here: " + s);
	}

	public final SJSocket receiveSession() throws SJIOException
	{
		throw new RuntimeException("Shouldn't get in here.");
	}

	public final SJServerAddress receiveServerAddress()
	{
		throw new RuntimeException("Shouldn't get in here.");
	}

	public final void recurse(String lab)
	{
		throw new RuntimeException("Shouldn't get in here: " + lab);
	}

	public final Object receive(int timeout) throws SJIOException, ClassNotFoundException
	{
		Socket s = getSocket();
		Object obj = null;

		try
		{
			int prev = s.getSoTimeout();

			s.setSoTimeout(timeout);

			obj = receive();

			s.setSoTimeout(prev);
		}
		catch (SocketException se)
		{
			throw new SJIOException(se);
		}

		return obj;
	}

	public final boolean receiveBoolean(int timeout) throws SJIOException
	{
		Socket s = getSocket();
		boolean v = false;

		try
		{
			int prev = s.getSoTimeout();

			s.setSoTimeout(timeout);

			v = receiveBoolean();

			s.setSoTimeout(prev);
		}
		catch (SocketException se)
		{
			throw new SJIOException(se);
		}

		return v;
	}

	public final int receiveInt(int timeout) throws SJIOException
	{
		Socket s = getSocket();
		int v = 0;

		try
		{
			int prev = s.getSoTimeout();

			s.setSoTimeout(timeout);

			v = receiveInt();

			s.setSoTimeout(prev);
		}
		catch (SocketException se)
		{
			throw new SJIOException(se);
		}

		return v;
	}

	public final byte receiveByte(int timeout) throws SJIOException
	{
		Socket s = getSocket();
		byte v = 0;

		try
		{
			int prev = s.getSoTimeout();

			s.setSoTimeout(timeout);

			v = receiveByte();

			s.setSoTimeout(prev);
		}
		catch (SocketException se)
		{
			throw new SJIOException(se);
		}

		return v;
	}

	public final SJSocket receiveSession(int timeout, String encoded) throws SJIOException
	{
		Socket s = getSocket();
		SJSocket v = null;

		try
		{
			int prev = s.getSoTimeout();

			s.setSoTimeout(timeout);

			v = receiveSession(encoded);

			s.setSoTimeout(prev);
		}
		catch (SocketException se)
		{
			throw new SJIOException(se);
		}

		return v;
	}

	public final SJServerAddress receiveServerAddress(int timeout, String encoded) throws SJIOException
	{
		Socket s = getSocket();
		SJServerAddress c = null;

		try
		{
			int prev = s.getSoTimeout();

			s.setSoTimeout(timeout);

			c = receiveServerAddress(encoded);

			s.setSoTimeout(prev);
		}
		catch (SocketException se)
		{
			throw new SJIOException(se);
		}

		return c;
	}

	public final String inlabel(int timeout) throws SJIOException, ClassNotFoundException
	{
		Socket s = getSocket();
		String lab = null;

		try
		{
			int prev = s.getSoTimeout();

			s.setSoTimeout(timeout);

			lab = inlabel();

			s.setSoTimeout(prev);
		}
		catch (SocketException se)
		{
			throw new SJIOException(se);
		}

		return lab;
	}

	public final boolean insync(int timeout) throws SJIOException
	{
		Socket s = getSocket();
		boolean v = false;

		try
		{
			int prev = s.getSoTimeout();

			s.setSoTimeout(timeout);

			v = insync();

			s.setSoTimeout(prev);
		}
		catch (SocketException se)
		{
			throw new SJIOException(se);
		}

		return v;
	}
}
